# This file is part of scrilla: https://github.com/chinchalinchin/scrilla.

# scrilla is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3
# as published by the Free Software Foundation.

# scrilla is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with scrilla.  If not, see <https://www.gnu.org/licenses/>
# or <https://github.com/chinchalinchin/scrilla/blob/develop/main/LICENSE>.

"""
# CLI Entrypoint 
This script acts as the entrypoint for the CLI application and contains the majority of the control structures for the program. It parses the arguments supplied through the command line, delegates them to the appropriate application function and then passes the results to the `scrilla.util.outputter` module functions for formatting and printing to screen.

The arguments are parsed in such a way that arguments which are not supplied are set to None. All application functions are set up to accept None as a value for their optional arguments. This makes passing arguments to application functions easier as the `main.py` script doesn't have to worry about their values. In other words, `main.py` always passes all arguments to application functions, even if they aren't supplied through the command line; it just sets the ones which aren't supplied to None.
"""

import sys, os, json
from typing import Callable

from scrilla import settings, services, files
from scrilla.static import functions, definitions, formats, keys
from scrilla.errors import InputValidationError
from scrilla.util import dater, outputter, helper
from scrilla.analysis import optimizer, markets, estimators
from scrilla.analysis.models.geometric import statistics, probability

# TODO: conditional imports based on value of ANALYSIS_MODE

from scrilla.analysis.objects.portfolio import Portfolio
from scrilla.analysis.objects.cashflow import Cashflow

if settings.APP_ENV != "container":
    import scrilla.util.plotter as plotter

logger = outputter.Logger('main', settings.LOG_LEVEL)

non_container_functions = [definitions.FUNC_DICT['plot_dividends'], definitions.FUNC_DICT['plot_moving_averages'],
                               definitions.FUNC_DICT['plot_risk_profile'], definitions.FUNC_DICT['plot_frontier']]

def validate_function_usage(selection: str, args: list, wrapper_function: Callable[[], None], required_length: int=1, exact: bool=False) -> None:
    """
    Parameters
    ----------
    1. **selection** : ``str``
    2. **args**: ``list``
        List of ticker/statistic symbols whose length is be to validated.
    3. **wrapper_function** : ``func()``
    4. **required_length** : ``int``
        The number of tickers required for the function.
    5. **exact** : ``bool``
        *Optional*. If the required length constraint is an equality, set to `True`. If the constraint is an inequality, set to `False`. Defaults to `False`. 
    """
    if selection in non_container_functions and settings.APP_ENV == 'container':
        raise InputValidationError('Graphics functionality disabled when application is containerized.')

    if(not exact and (len(args)>(required_length-1))):
        wrapper_function()
    elif(exact and (len(args)==required_length)):
        wrapper_function()
    else:
        raise InputValidationError(f'Invalid number of arguments for \'{selection}\' function.')
        

def print_format_to_screen(args: dict) -> bool:
    """
    Checks if the inputted optional arguments allow printing pretty formatted text to screen.

    Parameters
    ----------
    1. **args** : ``dict``
        Formatted dictionary containing optional arguments. Result of a call to `helper.format_xtra_args_dict`.
    
    Returns
    -------
    ``bool``
    """
    return not args['json'] and not args['suppress_output']

def print_json_to_screen(args: dict) -> bool:
    """
    Checks if inputted optional arguments allow printing json formatted text to screen.
    
    Parameters
    ----------
    1. **args** : ``dict``
        Formatted dictionary containing optional arguments. Result of a call to `helper.format_xtra_args_dict`.
    
    Returns
    -------
    ``bool``
    """
    return args['json'] and not args['suppress_output']

def do_program() -> None:
    """
    Parses command line arguments and passes the formatted arguments to appropriate function from the library.
    """
    files.init_static_data()
    args = helper.format_args(sys.argv[1:], settings.ESTIMATION_METHOD)
    exact, selected_function = False, None

    ########## START CLI FUNCTION DEFINITIONS

    ####### NO ARGUMENT FUNCTIONS
    ### FUNCTION: Help Message
    if args['function_arg'] in definitions.FUNC_DICT["help"]['values']:
        def cli_help():
            outputter.help_msg()
        selected_function, required_length = cli_help, 0

    ### FUNCTION: Clear Cache
    elif args['function_arg'] in definitions.FUNC_DICT["clear_cache"]['values']:
        def cli_clear_cache():
            logger.comment(f'Clearing {settings.CACHE_DIR}')
            files.clear_directory(directory=settings.CACHE_DIR, retain=True)
        selected_function, required_length = cli_clear_cache, 0

    ### FUNCTION: Clear Static
    elif args['function_arg'] in definitions.FUNC_DICT["clear_static"]['values']:
        def cli_clear_static():
            logger.comment(f'Clearing {settings.STATIC_DIR}')
            files.clear_directory(directory=settings.STATIC_DIR, retain=True)
        selected_function, required_length = cli_clear_static, 0

    ### FUNCTION: Clear Common
    elif args['function_arg'] in definitions.FUNC_DICT["clear_common"]['values']:
        def cli_clear_common():
            logger.comment(f'Clearing {settings.COMMON_DIR}')
            files.clear_directory(directory=settings.COMMON_DIR, retain=True)
        selected_function, required_length = cli_clear_common, 0

    ### FUNCTION: Print Stock Watchlist
    elif args['function_arg'] in definitions.FUNC_DICT['list_watchlist']['values']:
        def cli_watchlist():
            tickers = files.get_watchlist()
            outputter.title_line("Stock Watchlist")
            outputter.print_list(tickers)
        selected_function, required_length = cli_watchlist, 0

    ### FUNCTION: Purge Data Directories
    elif args['function_arg'] in definitions.FUNC_DICT["purge"]['values']:
        def cli_purge():
            logger.comment(f'Clearing {settings.STATIC_DIR}, {settings.CACHE_DIR} and {settings.CACHE_DIR}')
            files.clear_directory(directory=settings.STATIC_DIR, retain=True)
            files.clear_directory(directory=settings.CACHE_DIR, retain=True)
            files.clear_directory(directory=settings.COMMON_DIR, retain=True)
        selected_function, required_length = cli_purge, 0

    ### FUNCTION: Display Version
    elif args['function_arg'] in definitions.FUNC_DICT["version"]['values']:
        def cli_version():
            version_file = os.path.join(settings.APP_DIR, 'version.txt')
            with open(version_file, 'r') as f:
                print(f.read())
        selected_function, required_length = cli_version, 0

    ### FUNCTION: Yield Curve
    elif args['function_arg'] in definitions.FUNC_DICT['yield_curve']['values']:
        def cli_yield_curve():
            yield_curve = {}
            for maturity in keys.keys['YIELD_CURVE']:
                curve_rate = services.get_daily_interest_latest(maturity=maturity)
                yield_curve[maturity] = curve_rate/100

                if print_format_to_screen(args):
                    outputter.scalar_result(calculation=maturity, result=curve_rate/100, currency=False)

            if print_json_to_screen(args):
                print(json.dumps(yield_curve))
            
            if args['save_file'] is not None:
                files.save_file(file_to_save=yield_curve, file_name=args['save_file'])
        selected_function, required_length = cli_yield_curve, 0

    ####### ARGUMENT FUNCTIONS
    ### FUNCTION: Asset Grouping

    elif args['function_arg'] in definitions.FUNC_DICT['asset_type']['values']:
        def cli_asset_type():
            for arg in args['tickers']:
                asset_type = files.get_asset_type(arg)
                outputter.string_result(f'asset_type({arg})', asset_type)
        selected_function, required_length = cli_asset_type, 1

    ### FUNCTION: Black-Scholes Value At Risk
    elif args['function_arg'] in definitions.FUNC_DICT['var']['values']:
        def cli_var():
            all_vars = {}
            for arg in args['tickers']['values']:
                prices = services.get_daily_price_history(ticker=arg, 
                                                            start_date=args['start_date'],
                                                            end_date=args['end_date'])
                latest_price = prices[helper.get_first_json_key(prices)][keys.keys['PRICES']['CLOSE']]
                profile = statistics.calculate_risk_return(ticker=arg, 
                                                            sample_prices=prices, 
                                                            method=args['estimation_method'])
                valueatrisk = probability.percentile(S0=latest_price, 
                                                        vol=profile['annual_volatility'],
                                                        ret=profile['annual_return'], 
                                                        expiry=args['expiry'],
                                                        percentile=args['probability'])
                all_vars[arg]=valueatrisk

                if print_format_to_screen(args):
                    outputter.scalar_result(f'{arg}_value_at_risk', valueatrisk)

            if print_json_to_screen(args):
                print(json.dumps(all_vars))
            
            if args['save_file'] is not None:
                files.save_file(file_to_save=all_vars, file_name=args['save_file'])

        selected_function, required_length = cli_var, 2
        
    ### FUNCTION: Black-Scholes Conditional Value At Risk
    elif args['function_arg'] in definitions.FUNC_DICT['cvar']['values']:
        def cli_cvar():
            all_cvars = {}
            for arg in args['tickers']:
                prices = services.get_daily_price_history(ticker=arg, 
                                                            start_date=args['start_date'],
                                                            end_date=args['end_date'])
                latest_price = prices[helper.get_first_json_key(prices)][keys.keys['PRICES']['CLOSE']]
                profile = statistics.calculate_risk_return(ticker=arg, 
                                                            sample_prices=prices, 
                                                            method=args['estimation_method'])
                valueatrisk = probability.percentile(S0=latest_price, 
                                                        vol=profile['annual_volatility'],
                                                        ret=profile['annual_return'], 
                                                        expiry=args['expiry'],
                                                        percentile=args['probability'])
                cvar = probability.conditional_expected_value(S0=latest_price, 
                                                                vol=profile['annual_volatility'],
                                                                ret=profile['annual_return'], 
                                                                expiry=args['expiry'],
                                                                conditional_value=valueatrisk)
                all_cvars[arg]=cvar

                if print_format_to_screen(args):
                    outputter.scalar_result(f'{arg}_conditional_value_at_risk', valueatrisk)

            if print_json_to_screen(args):
                print(json.dumps(all_cvars))
            
            if args['save_file'] is not None:
                files.save_file(file_to_save=all_cvars, file_name=args['save_file'])

        selected_function, required_length = cli_cvar, 2

    ### FUNCTION: Capital Asset Pricing Model Cost of Equity
    elif args['function_arg'] in definitions.FUNC_DICT['capm_equity_cost']['values']:
        def cli_capm_equity_cost():
            all_costs = {}
            for arg in args['tickers']:
                equity_cost = markets.cost_of_equity(ticker=arg, start_date=args['start_date'], 
                                                        end_date=args['end_date'], 
                                                        method=args['estimation_method'])
                all_costs[arg] = equity_cost

                if print_format_to_screen(args):
                    outputter.scalar_result(f'{arg}_equity_cost', equity_cost, currency=False)

            if print_json_to_screen(args):
                print(json.dumps(all_costs))

            if args['save_file'] is not None:
                files.save_file(file_to_save=all_costs, file_name=args['save_file'])

        selected_function, required_length = cli_capm_equity_cost, 1


    ### FUNCTION: Capital Asset Pricing Model Beta
    elif args['function_arg'] in definitions.FUNC_DICT['capm_beta']['values']:
        def cli_capm_beta():
            all_betas = {}
            for arg in args['tickers']:
                beta = markets.market_beta(ticker=arg, 
                                            start_date=args['start_date'], 
                                            end_date=args['end_date'], 
                                            method=args['estimation_method'])
                all_betas[arg] = beta

                if print_format_to_screen(args):
                    outputter.scalar_result(f'{arg}_beta', beta, currency=False)

            if print_json_to_screen(args):
                print(json.loads(all_betas))

            if args['save_file'] is not None:
                files.save_file(file_to_save=all_betas, file_name=args['save_file'])
        selected_function, required_length = cli_capm_beta, 1

    ### FUNCTION: Last Close Price
    elif args['function_arg'] in definitions.FUNC_DICT["close"]['values']:
        def cli_close():
            all_prices = {}
            for arg in args['tickers']:
                price = services.get_daily_price_latest(arg)
                all_prices[arg] = price

                if print_format_to_screen(args):
                    outputter.scalar_result(calculation=f'Last {arg} close price', result=float(price))

            if print_json_to_screen(args):
                print(json.dumps(all_prices))

            if args['save_file'] is not None:
                files.save_file(file_to_save=all_prices, file_name=args['save_file'])
        selected_function, required_length = cli_close, 1
            
    ### FUNCTION: Correlation Matrix
    elif args['function_arg'] in definitions.FUNC_DICT["correlation"]['values']:
        def cli_correlation():
            if args['estimation_method'] == keys.keys['ESTIMATION']['LIKE']:
                logger.comment('This calculation takes a while, strap in...')

            matrix = statistics.correlation_matrix(tickers=args['tickers'],
                                                    start_date=args['start_date'], 
                                                    end_date=args['end_date'],
                                                    method=args['estimation_method'])
            if print_format_to_screen(args):
                outputter.correlation_matrix(tickers=args['tickers'], correlation_matrix=matrix)

            elif print_json_to_screen(args):
                print(json.dumps(matrix))

            if args['save_file'] is not None:
                files.save_file(file_to_save=matrix, file_name=args['save_file'])
                
        selected_function, required_length = cli_correlation, 2

    ### FUNCTION: Correlation Time Series
    elif args['function_arg'] in definitions.FUNC_DICT['correlation_time_series']['values']:
        def cli_correlation_series():
            logger.comment('This calculation takes a while, strap in...')
            ticker_1, ticker_2 = args['tickers'][0], args['tickers'][1]
            result = statistics.calculate_moment_correlation_series(ticker_1=ticker_1,ticker_2=ticker_2,
                                                                    start_date=args['start_date'],
                                                                    end_date=args['end_date'])
            if print_format_to_screen(args):
                for date in result:
                    outputter.scalar_result(calculation=f'{date}_{ticker_1}_{ticker_2}_correlation', 
                                        result=float(result[date]), currency=False)
            elif print_json_to_screen(args):
                print(json.dumps(result))

            if args['save_file'] is not None:
                files.save_file(file_to_save=result, file_name=args['save_file'])
        selected_function, required_length, exact = cli_correlation_series, 2, True

    ### FUNCTION: Discount Dividend Model
    elif args['function_arg'] in definitions.FUNC_DICT["discount_dividend"]['values']:
        def cli_discount_dividend():
            model_results = {}
            for arg in args['tickers']:
                dividends = services.get_dividend_history(arg)
                if args['discount'] is None:
                    discount = markets.cost_of_equity(ticker=arg, method=args['estimation_method'])
                else:
                    discount = args['discount']
                model_results[f'{arg}_discount_dividend'] = Cashflow(sample=dividends, 
                                                                        discount_rate=discount).calculate_net_present_value()
                if print_format_to_screen(args):
                    outputter.scalar_result(f'Net Present Value ({arg} dividends)', 
                                            model_results[f'{arg}_discount_dividend'])

            if print_json_to_screen(args):
                print(json.dumps(model_results))

            if args['save_file'] is not None:
                files.save_file(file_to_save=model_results, file_name=args['save_file'])
        selected_function, required_length = cli_discount_dividend, 1

    elif args['function_arg'] in definitions.FUNC_DICT['dividends']['values']:
        def cli_dividends():
            all_dividends = {}
            for arg in args['tickers']:
                dividends = services.get_dividend_history(arg)
                all_dividends[arg] = dividends
                if print_format_to_screen(args):
                    for date in dividends:
                        outputter.scalar_result(calculation=f'{arg}_dividend({date})', result=dividends[date])
        
            if print_json_to_screen(args):
                print(json.dumps(all_dividends))
        selected_function, required_length = cli_dividends, 1

    ### FUNCTION: Efficient Frontier
    elif args['function_arg'] in definitions.FUNC_DICT['efficient_frontier']['values']:
        def cli_efficient_frontier():
            portfolio = Portfolio(tickers=args['tickers'], 
                                    start_date=args['start_date'], 
                                    end_date=args['end_date'],
                                    method=args['estimation_method'])
            frontier = optimizer.calculate_efficient_frontier(portfolio=portfolio, 
                                                                steps=args['steps'])

            if print_format_to_screen(args):
                outputter.efficient_frontier(portfolio=portfolio, 
                                                    frontier=frontier,
                                                    investment=args['investment'])
            if print_json_to_screen(args):
                print(json.dumps(functions.format_frontier(portfolio=portfolio, 
                                                                frontier=frontier, 
                                                                investment=args['investment'])))

            if args['save_file'] is not None:
                files.save_frontier(portfolio=portfolio, 
                                    frontier=frontier,
                                    investment=args['investment'], 
                                    file_name=args['save_file'])
        selected_function, required_length = cli_efficient_frontier, 2

    ### FUNCTION: Maximize Portfolio Return
    elif args['function_arg'] in definitions.FUNC_DICT['maximize_return']['values']:
        def cli_maximize_return():
            portfolio = Portfolio(tickers=args['tickers'], 
                                    start_date=args['start_date'], 
                                    end_date=args['end_date'],
                                    method=args['estimation_method'])
            allocation = optimizer.maximize_portfolio_return(portfolio=portfolio)

            if print_format_to_screen(args):
                outputter.optimal_result(portfolio=portfolio, 
                                                allocation=allocation, 
                                                investment=args['investment'])
            if print_json_to_screen(args):
                print(json.dumps(functions.format_allocation(allocation=allocation, 
                                                                    portfolio=portfolio, 
                                                                    investment=args['investment'])))

            if args['save_file'] is not None:
                files.save_allocation(allocation=allocation, 
                                        portfolio=portfolio, 
                                        file_name=args['save_file'], 
                                        investment=args['investment'])
        selected_function, required_length = cli_maximize_return, 2

    ### FUNCTION: Moving Averages of Logarithmic Returns
    elif args['function_arg'] in definitions.FUNC_DICT['moving_averages']['values']:
        def cli_moving_averages():
            # TODO: moving averages with estimation techniques
            # TODO: print results as json to screen and ability to save results
            moving_averages = statistics.calculate_moving_averages(tickers=args['tickers'], 
                                                                    start_date=args['start_date'], 
                                                                    end_date=args['end_date'])
            periods = [settings.MA_1_PERIOD, settings.MA_2_PERIOD, settings.MA_3_PERIOD]

            outputter.moving_average_result(tickers=args['tickers'], averages_output=moving_averages, 
                                            periods=periods, start_date=args['start_date'], 
                                            end_date=args['end_date'])
        selected_function, required_length = cli_moving_averages, 1

    ### FUNCTION: Optimize Portfolio Variance/Volatility
    elif args['function_arg'] in definitions.FUNC_DICT['optimize_portfolio']['values']:
        def cli_optimize_portfolio_variance():
            portfolio = Portfolio(tickers=args['tickers'], 
                                    start_date=args['start_date'], 
                                    end_date=args['end_date'],
                                    method=args['estimation_method'])

            if args['optimize_sharpe']:
                allocation = optimizer.maximize_sharpe_ratio(portfolio=portfolio, target_return=args['target'])
            else:
                allocation = optimizer.optimize_portfolio_variance(portfolio=portfolio, target_return=args['target'])   
            
            if print_format_to_screen(args):
                outputter.optimal_result(portfolio=portfolio, allocation=allocation, investment=args['investment'])

            if print_json_to_screen(args):
                print(json.dumps(functions.format_allocation(allocation=allocation,portfolio=portfolio, investment=args['investment'])))
            
            if args['save_file'] is not None:
                files.save_allocation(allocation=allocation, portfolio=portfolio, file_name=args['save_file'],
                                        investment=args['investment'])
        selected_function, required_length = cli_optimize_portfolio_variance, 2
    
    ### FUNCTION: Optimize Portfolio Conditional Value At Risk 
    elif args['function_arg'] in definitions.FUNC_DICT['optimize_portfolio_conditional_var']['values']:
        def cli_optimize_conditional_value_at_risk():
            portfolio = Portfolio(tickers=args['tickers'], 
                                    start_date=args['start_date'], 
                                    end_date=args['end_date'],
                                    method=args['estimation_method'])
            allocation = optimizer.optimize_conditional_value_at_risk(portfolio=portfolio,
                                                                        prob=args['probability'],
                                                                        expiry=args['expiry'],
                                                                        target_return=args['target'])
            if print_format_to_screen(args):
                outputter.optimal_result(portfolio=portfolio, allocation=allocation, investment=args['investment'])

            if print_json_to_screen(args):
                print(json.dumps(functions.format_allocation(allocation=allocation, portfolio=portfolio, investment=args['investment'])))
            
            if args['save_file'] is not None:
                files.save_allocation(allocation=allocation, portfolio=portfolio, file_name=args['save_file'],
                                        investment=args['investment'])
        selected_function, required_length = cli_optimize_conditional_value_at_risk, 2

    ### FUNCTION: Plot Correlation Time Series
    elif args['function_arg'] in definitions.FUNC_DICT['plot_correlation']['values']:
        def cli_plot_correlation():
            logger.comment('This calculation takes a while, strap in...')
            correlation_history = statistics.calculate_moment_correlation_series(ticker_1=args['tickers'][0], 
                                                                                ticker_2=args['tickers'][1], 
                                                                                start_date=args['start_date'],
                                                                                end_date=args['end_date'])
            plotter.plot_correlation_series(tickers=args['tickers'], series=correlation_history, savefile=args['save_file'])
        selected_function, required_length, exact = cli_plot_correlation, 2, True

    ### FUNCTION: Plot Dividend History With Linear Regression Model
    elif args['function_arg'] in definitions.FUNC_DICT['plot_dividends']['values']:
        def cli_plot_dividends():
            dividends = services.get_dividend_history(ticker=args['tickers'][0])
            if args['discount'] is None:
                args['discount'] = markets.cost_of_equity(ticker=args['tickers'][0],
                                                                method=args['estimation_method'])
            div_cashflow = Cashflow(sample=dividends, 
                                    discount_rate=args['discount'])
            plotter.plot_cashflow(ticker=args['tickers'][0], 
                                    cashflow=div_cashflow, show=True, 
                                    savefile=args['save_file'])
        selected_function, required_length, exact = cli_plot_dividends, 1, True

    ### FUNCTION: Plot Efficient Frontier
    elif args['function_arg'] in definitions.FUNC_DICT['plot_frontier']['values']:
        def cli_plot_frontier():
            portfolio = Portfolio(tickers=args['tickers'], 
                                    start_date=args['start_date'], 
                                    end_date=args['end_date'],
                                    method=args['estimation_method'])
            frontier = optimizer.calculate_efficient_frontier(portfolio=portfolio,
                                                                steps=args['steps'])
            plotter.plot_frontier(portfolio=Portfolio(args['tickers']), 
                                    frontier=frontier, 
                                    show=True, 
                                    savefile=args['save_file'])
        selected_function, required_length = cli_plot_frontier, 2

    ### FUNCTION: Plot Moving Averages of Logarithmic Returns
    elif args['function_arg'] in definitions.FUNC_DICT['plot_moving_averages']['values']:
        def cli_plot_moving_averages():
            # TODO: estimation techniques with moving averages
            moving_averages = statistics.calculate_moving_averages(tickers=args['tickers'], 
                                                                    start_date=args['start_date'], 
                                                                    end_date=args['end_date'])
            periods = [settings.MA_1_PERIOD, settings.MA_2_PERIOD, settings.MA_3_PERIOD]
            plotter.plot_moving_averages(symbols=args['tickers'], 
                                            averages_output=moving_averages, 
                                            periods=periods, 
                                            show=True, savefile=args['save_file'])
        selected_function, required_length = cli_plot_moving_averages, 1

    ### FUNCTION: Plot Return QQ Series
    elif args['function_arg'] in definitions.FUNC_DICT['plot_returns']['values']:
        def cli_plot_returns():
            asset_type = files.get_asset_type(symbol=args['tickers'][0])
            prices = services.get_daily_price_history(ticker=args['tickers'][0], 
                                                        start_date=args['start_date'],
                                                        end_date=args['end_date'],
                                                        asset_type=asset_type)
            returns = statistics.get_sample_of_returns(prices, asset_type)
            qq_series = estimators.qq_series_for_sample(sample=returns)
            plotter.plot_qq_series(ticker=args['tickers'][0], 
                                    sample=qq_series, 
                                    show=True, 
                                    savefile=args['save_file'])
        selected_function, required_length, exact = cli_plot_returns, 1, True

    ### FUNCTION: Plot Risk-Return Profile
    elif args['function_arg'] in definitions.FUNC_DICT['plot_risk_profile']['values']:
        def cli_plot_risk_profile():
            profiles = {}
            for arg in args['tickers']:
                profiles[arg]=statistics.calculate_risk_return(ticker=arg, 
                                                                start_date=args['start_date'], 
                                                                end_date=args['end_date'],
                                                                method=args['estimation_method'])
            plotter.plot_profiles(symbols=args['tickers'], 
                                    show=True,
                                    profiles=profiles, 
                                    savefile=args['save_file'], 
                                    subtitle=dater.format_date_range(start_date=args['start_date'], 
                                                                        end_date=args['end_date']))
        selected_function, required_length = cli_plot_risk_profile, 1

    elif args['function_arg'] in definitions.FUNC_DICT['plot_yield_curve']['values']:
        def cli_plot_yield_curve():
            yield_curve = {}
            args['start_date'] = dater.get_next_business_date(args['start_date'])
            start_date_string = dater.date_to_string(args['start_date'])
            yield_curve[start_date_string] = []
            for maturity in keys.keys['YIELD_CURVE']:
                rate = services.get_daily_interest_history(maturity=maturity, 
                                                            start_date=args['start_date'],
                                                            end_date=args['start_date'])
                yield_curve[start_date_string].append(rate[start_date_string])
        
            plotter.plot_yield_curve(yield_curve=yield_curve, show=True,
                                        savefile=args['save_file'])
        selected_function, required_length, exact = cli_plot_yield_curve, 0, True
    ### FUNCTION: Price History
    elif args['function_arg'] in definitions.FUNC_DICT['price_history']['values']:
        def cli_price_history():
            all_prices = {}
            for arg in args['tickers']:
                prices = services.get_daily_price_history(ticker=arg, 
                                                            start_date=args['start_date'],
                                                            end_date=args['end_date'])
                all_prices[arg] = {}
                for date in prices:
                    price = prices[date][keys.keys['PRICES']['CLOSE']]
                    all_prices[arg][date] = price

                    if print_format_to_screen(args):
                        outputter.scalar_result(calculation=f'{arg}({date})', result = float(price))

            if print_json_to_screen(args):
                print(json.dumps(all_prices))

            if args['save_file'] is not None:
                files.save_file(file_to_save=all_prices, file_name=args['save_file'])
        selected_function, required_length = cli_price_history, 1

    ### FUNCTION: Interest Rate History
    elif args['function_arg'] in definitions.FUNC_DICT['interest_history']['values']:
        def cli_interest_history():
            all_rates = {}
            for arg in args['tickers']:
                all_rates[arg] = services.get_daily_interest_history(maturity=arg, 
                                                                        start_date=args['start_date'],
                                                                        end_date=args['end_date'])
                for date in all_rates[arg]:
                    all_rates[arg][date] = all_rates[arg][date]/100
                    
                if print_format_to_screen(args):
                    for date in all_rates[arg]:
                        outputter.scalar_result(calculation=f'{arg}_YIELD({date})', result=float(all_rates[arg][date])/100, currency=False)

            if print_json_to_screen(args):
                print(json.dumps(all_rates))

            if args['save_file'] is not None:
                files.save_file(file_to_save=all_rates, file_name=args['save_file'])
        selected_function, required_length = cli_interest_history, 1
    
    ### FUNCTION: Risk Free Rate
    elif args['function_arg'] in definitions.FUNC_DICT['risk_free_rate']['values']:
        def cli_risk_free_rate():
            rate = {}
            rate[settings.RISK_FREE_RATE] = services.get_risk_free_rate()

            if print_format_to_screen(args):
                    outputter.title_line("Risk Free Rate")
                    outputter.scalar_result(calculation=formats.formats.RISK_FREE_TITLE.format(settings.RISK_FREE_RATE), 
                                                result=rate[settings.RISK_FREE_RATE], currency=False)
            if print_json_to_screen:
                print(json.dumps(rate))

            if args['save_file'] is not None:
                files.save_file(file_to_save=rate, file_name=args['save_file'])
        selected_function, required_length, exact = cli_risk_free_rate, 0, True

    ### FUNCTION: Risk-Return Profile
    elif args['function_arg'] in definitions.FUNC_DICT["risk_profile"]['values']:
        def cli_risk_return():
            profiles = {}
            for arg in args['tickers']:
                profiles[arg] = statistics.calculate_risk_return(ticker=arg, 
                                                                    method = args['estimation_method'],
                                                                    start_date=args['start_date'], 
                                                                    end_date=args['end_date'])
                profiles[arg]['sharpe_ratio'] = markets.sharpe_ratio(ticker=arg, 
                                                                    start_date=args['start_date'],
                                                                    end_date=args['end_date'], 
                                                                    ticker_profile=profiles[arg],
                                                                    method=args['estimation_method'])
                profiles[arg]['asset_beta'] = markets.market_beta(ticker=arg, 
                                                                    start_date=args['start_date'],
                                                                    end_date=args['end_date'],
                                                                    ticker_profile=profiles[arg],
                                                                    method=args['estimation_method'])
                profiles[arg]['equity_cost'] = markets.cost_of_equity(ticker=arg, 
                                                                    start_date=args['start_date'],
                                                                    end_date=args['end_date'], 
                                                                    method=args['estimation_method'])
            
            if print_format_to_screen(args):
                outputter.risk_profile(profiles=profiles)

            if print_json_to_screen(args):
                print(json.dumps(profiles))

            if args['save_file'] is not None:
                files.save_profiles(profiles=profiles, file_name=args['save_file'])
        selected_function, required_length = cli_risk_return, 1

    ### FUNCTION: Model Discount Screener 
    elif args['function_arg'] in definitions.FUNC_DICT["screener"]['values']:
        def cli_screener():
            if args['model'] is None:
                model = markets.MODEL_DDM
            results = markets.screen_for_discount(model=model, discount_rate=args['discount'])
            outputter.screen_results(info=results, model=model)
        selected_function, required_length = cli_screener, 0

    ### FUNCTION: Sharpe Ratio
    elif args['function_arg'] in definitions.FUNC_DICT["sharpe_ratio"]['values']:
        def cli_sharpe_ratio():
            all_results = {}
            for arg in args['tickers']:
                result = markets.sharpe_ratio(ticker=arg, 
                                                start_date=args['start_date'],
                                                end_date=args['end_date'],
                                                method=args['estimation_method'])
                all_results[arg]=result
                
                if print_format_to_screen(args):
                    outputter.scalar_result(calculation=f'{arg}_sharpe_ratio', result=result, 
                                            currency=False)

            if print_json_to_screen(args):
                print(json.dumps(all_results))

            if args['save_file'] is not None:
                files.save_file(file_to_save=all_results, file_name=args['save_file'])
        selected_function, required_length = cli_sharpe_ratio, 1

   ### FUNCTION: Store Key
    elif args['function_arg'] in definitions.FUNC_DICT['store']['values']:
        def cli_store():
            files.set_credentials(value=args['value'], which_key=args['key'])
        selected_function, required_length = cli_store, 0

    ### FUNCTION: Get Latest Economic Statistic
    elif args['function_arg'] in definitions.FUNC_DICT["statistic"]['values']:
        def cli_statistic():
            all_stats = {}
            for stat in args['tickers']:
                result = services.get_daily_fred_latest(symbol=stat)
                all_stats[stat] = result
                if print_format_to_screen(args):
                    outputter.scalar_result(calculation=stat, result=result, currency=False)

            if print_json_to_screen(args):
                print(json.dumps(all_stats))

            if args['save_file'] is not None:
                files.save_file(file_to_save=all_stats, file_name=args['save_file'])
        selected_function, required_length = cli_statistic, 1
    
    ### FUNCTION: Statistic History
    elif args['function_arg'] in definitions.FUNC_DICT['statistic_history']['values']:
        def cli_statistic_history():
            all_stats = {}
            for arg in args['tickers']:
                stats = services.get_daily_fred_history(symbol=arg, 
                                                        start_date=args['start_date'],
                                                        end_date=args['end_date'])
                all_stats[arg] = stats
                if print_format_to_screen(args):
                    for date in stats:
                        outputter.scalar_result(calculation=f'{arg}({date})', 
                                                    result=stats[date], 
                                                    currency=False) 

            if print_json_to_screen(args):
                print(json.dumps(all_stats))
            if args['save_file'] is not None:
                files.save_file(file_to_save=all_stats, file_name=args['save_file'])
        selected_function, required_length = cli_statistic_history, 1

    ### FUNCTION: Set Watchlist
    elif args['function_arg'] in definitions.FUNC_DICT["watchlist"]['values']:
        def cli_watchlist():
            files.add_watchlist(new_tickers=args['tickers'])
            logger.comment("Watchlist saved. Use -ls option to print watchlist.")
        selected_function, required_length = cli_watchlist, 1

    else:
        def cli_help():
            outputter.help_msg()
        selected_function, required_length = cli_help, 0
        
    ########## END CLI FUNCTION DEFINITIONS
      
    if selected_function is not None:
        validate_function_usage(selection=args['function_arg'], 
                                    args=args['tickers'],
                                    wrapper_function=selected_function, 
                                    required_length=required_length, 
                                    exact=exact)

if __name__ == "__main__": 
    do_program()


# This file is part of scrilla: https://github.com/chinchalinchin/scrilla.

# scrilla is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3
# as published by the Free Software Foundation.

# scrilla is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with scrilla.  If not, see <https://www.gnu.org/licenses/>
# or <https://github.com/chinchalinchin/scrilla/blob/develop/main/LICENSE>.

import sys

from PySide6 import QtCore, QtWidgets, QtGui

from scrilla import settings

from scrilla.gui import formats
from scrilla.gui.widgets.functions import RiskReturnWidget, CorrelationWidget, \
                                            MovingAverageWidget, EfficientFrontierWidget, \
                                            OptimizerWidget

# NOTE: widget_buttons and function_widgets must preserve order.
class MenuWidget(QtWidgets.QWidget):
    """
    
    .. notes::
        * Widget Hierarchy: 

    """
    def __init__(self):
        super().__init__()
        self._init_menu_widgets()
        self._generate_menu_bar()
        self._arrange_menu_widgets()
        self._stage_menu_widgets()

    def _generate_menu_bar(self):
        self.api_key_action = QtGui.QAction('Add API Key', self)
        self.jpeg_export_action = QtGui.QAction('JPEG', self)
        self.json_export_action = QtGui.QAction('JSON', self)

        self.menu_bar = QtWidgets.QMenuBar(None)
        self.menu_bar.setObjectName('menu-bar')

        file = self.menu_bar.addMenu('File')
        view = self.menu_bar.addMenu('Function')
        preferences = self.menu_bar.addMenu('Preferences')

        self.function_actions = [ QtGui.QAction(function, self) for function in formats.FUNCTIONS ]
        for action in self.function_actions:
            view.addAction(action)

        export = file.addMenu('Export')

        self.api_key_action.setEnabled(True)
        self.jpeg_export_action.setEnabled(False)
        self.json_export_action.setEnabled(False)

        file.addAction(self.api_key_action)
        export.addAction(self.jpeg_export_action)
        export.addAction(self.json_export_action)

    def _init_menu_widgets(self):
        self.setObjectName('root')

        self.intro_msg = QtWidgets.QLabel("Select A Function To Get Started")
        self.intro_msg.setObjectName('title')

        self.title = QtWidgets.QLabel("scrilla")
        self.title.setObjectName('title')

        self.function_menu = QtWidgets.QLabel('Functions')
        self.function_menu.setObjectName('function-menu')
        
        
        self.widget_buttons = [ QtWidgets.QPushButton(function) for function in formats.FUNCTIONS ]

        self.exit_button = QtWidgets.QPushButton("Exit")
        self.exit_button.setObjectName('button')

        for button in self.widget_buttons:
            button.setObjectName('button')

        self.function_widgets = [ 
            CorrelationWidget('great-grand-child'), 
            EfficientFrontierWidget('great-grand-child'),
            MovingAverageWidget('great-grand-child'),
            OptimizerWidget('great-grand-child'),
            RiskReturnWidget('great-grand-child'),
        ]

        self.menu_pane = QtWidgets.QWidget()
        self.menu_pane.setLayout(QtWidgets.QVBoxLayout())
        self.menu_pane.setObjectName('grand-child')
        
        self.display_pane = QtWidgets.QWidget()
        self.display_pane.setLayout(QtWidgets.QVBoxLayout())
        self.display_pane.setObjectName('grand-child')

        self.container_pane = QtWidgets.QWidget()
        self.container_pane.setLayout(QtWidgets.QHBoxLayout())
        self.container_pane.setObjectName('child')

        self.setLayout(QtWidgets.QVBoxLayout())

    def _arrange_menu_widgets(self):
        """Arranges child widget within their layouts."""
        self.title.setAlignment(QtCore.Qt.AlignHCenter)
        self.intro_msg.setAlignment(QtCore.Qt.AlignCenter)
        self.function_menu.setAlignment(QtCore.Qt.AlignHCenter)

        self.container_pane.setSizePolicy(QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding,
                                                                QtWidgets.QSizePolicy.Expanding))
        self.menu_pane.setSizePolicy(QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum,
                                                            QtWidgets.QSizePolicy.Expanding))
        self.display_pane.setSizePolicy(QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding,
                                                            QtWidgets.QSizePolicy.Expanding))
        self.title.setSizePolicy(QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding,
                                                        QtWidgets.QSizePolicy.Minimum))

        self.container_pane.layout().addWidget(self.menu_pane)
        self.container_pane.layout().addWidget(self.display_pane)

        self.layout().addWidget(self.menu_bar)
        self.layout().addWidget(self.title)
        self.layout().addWidget(self.container_pane)

        self.menu_pane.layout().addWidget(self.function_menu)
        for button in self.widget_buttons:
            self.menu_pane.layout().addWidget(button)
        self.menu_pane.layout().addStretch()
        self.menu_pane.layout().addWidget(self.exit_button)

        self.display_pane.layout().addWidget(self.intro_msg)
        for widget in self.function_widgets:
            self.display_pane.layout().addWidget(widget)
        
    def _stage_menu_widgets(self):
        for i, button in enumerate(self.widget_buttons):
            button.show()
            button.setAutoDefault(True)
            button.clicked.connect((lambda i: lambda: self._show_widget(i))(i))
            button.setCursor(QtGui.QCursor(QtCore.Qt.PointingHandCursor))
        for widget in self.function_widgets:
            widget.hide()
        self.exit_button.clicked.connect(self.close)
        self.exit_button.show()
        self.menu_bar.show()

    @QtCore.Slot()
    def _show_widget(self, widget_index):
        self.intro_msg.hide()
        for widget in self.function_widgets:
            widget.hide()
        self.function_widgets[widget_index].show()

    @QtCore.Slot()
    def _clear(self):
        for widget in self.function_widgets:
            widget.hide()
        self.intro_msg.show()

if __name__ == "__main__":
    app = QtWidgets.QApplication([])

    mainWidget = MenuWidget()
    mainWidget.resize(settings.GUI_WIDTH, settings.GUI_HEIGHT)
    mainWidget.show()

    sys.exit(app.exec_())


# This file is part of scrilla: https://github.com/chinchalinchin/scrilla.

# scrilla is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3
# as published by the Free Software Foundation.

# scrilla is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with scrilla.  If not, see <https://www.gnu.org/licenses/>
# or <https://github.com/chinchalinchin/scrilla/blob/develop/main/LICENSE>.

from typing import Callable, Dict, List, Union
from PySide6 import QtGui, QtCore, QtWidgets

from scrilla.gui import utilities, formats
from scrilla.static import definitions, constants

# TODO: have widgets create their own layouts and arrange their hcild widgets in them, i.e. not in their root layout.
#       then compose the layouts in the functions module.

# in fact, if you think about it, none of these widgets need to inherit QWidget. they could implement Layouts and add their
# child widgets directly to their layouts. then in the functions module, the layouts could be composed...

# that's probably the way to do this...

# this is definitely not a clean way of doing. all the classes are mixed together. need to enforce separation or streamline
# the inheritance.

# question: where should the calculate buttons come from?
# question: what widgets are absolutely necessary for each widget in here?
# isolate by functionality. no widget dependency.

# REFACTOR.

# I think the calculate button and clear button coming from the symbolwidget is fine, but table/composite/graph/portfolio 
# should not hook up the functions to these buttons.

# no inheritance between widgets. 

"""
A series of classes that all inherit from ``PySide6.QtWidgets.QWidget` and build up in sequence the necessary functionality to grab and validate user input, calculate and display results, etc. These widgets are not directly displayed on the GUI; rather, they are used as building blocks by the `scrilla.gui.functions` module, to create widgets specifically for application functions.

All widgets have a similar structure in that they call a series of methods tailored to their specific class: 
    1. **_init**
        Child widgets are constructed. Layouts are created.
    2. **_arrange**
        Child widgets are arranged. Layouts are applied.
    3. **_stage**
        Child widgets are prepped for display.
"""


def generate_control_skeleton():
    return { arg: False for arg in definitions.ARG_DICT }

def _layout_factory(format: str):
    widget = QtWidgets.QWidget()

    if format == 'vertical-box':
        widget.setLayout(QtWidgets.QVBoxLayout())
    
    elif format == 'horizontal-box':
        widget.setLayout(QtWidgets.QHBoxLayout())
    
    else:
        widget.setLayout(QtWidgets.QBoxLayout())

    return widget

def _atomic_widget_factory(format: str, title: str):
    if format in ['title', 'subtitle', 'heading', 'label', 'error', 'text'] :
        widget = QtWidgets.QLabel(title)
        widget.setObjectName(format)
        if format in ['title', 'subtitle', 'label', 'heading']:
            widget.setAlignment(QtCore.Qt.AlignTop)
    
    elif format == 'button':
        widget = QtWidgets.QPushButton(title)
        widget.setObjectName(format)

    else: 
        widget = QtWidgets.QWidget()
    
    return widget
# this methdo was separated into input and passive factory methods. combined them because i thought it made sense,
# however, this method is now getting a little complex. it might not hurt to simplify and refactor a bit.
def _composite_widget_factory(format: str, title: str = None) -> QtWidgets.QWidget:
    # input widget

    if format == 'date':
        widget = QtWidgets.QWidget()
        widget.setLayout(QtWidgets.QHBoxLayout())
        widget.setObjectName('input-container')

        label_widget = QtWidgets.QLabel(title)
        label_widget.setAlignment(QtCore.Qt.AlignBottom)
        label_widget.setObjectName('input-label')

        main_widget = QtWidgets.QDateEdit()
        main_widget.setDate(QtCore.QDate.currentDate())
        main_widget.setMaximumDate(QtCore.QDate.currentDate())
        main_widget.setMinimumDate(QtCore.QDate(constants.constants['PRICE_YEAR_CUTOFF'], 1,1))
        main_widget.setObjectName(format)

        widget.layout().addWidget(label_widget)
        widget.layout().addWidget(main_widget)

    elif format == 'decimal':
        pass

    elif format == 'currency':
        pass
    
    elif format == 'years':
        pass
    
    elif format == 'integer':
        pass
    
    elif format == 'model':
        pass
    
    elif format == 'select':
        pass
    
    elif format == 'file':
        pass
    
    elif format == 'key':
        pass
    
    elif format == 'password':
        pass
    
    elif format == 'flag':
        pass
    
    elif format == 'line-edit':
        widget = QtWidgets.QWidget()
        widget.setLayout(QtWidgets.QHBoxLayout())
        widget.setObjectName('input-container')

        label_widget = QtWidgets.QLabel(title)
        label_widget.setAlignment(QtCore.Qt.AlignBottom)
        label_widget.setObjectName('input-label')

        main_widget = QtWidgets.QLineEdit()
        main_widget.setAlignment(QtCore.Qt.AlignTop)
        main_widget.setObjectName(format)

        widget.layout().addWidget(label_widget)
        widget.layout().addWidget(main_widget)

    else:
        widget = QtWidgets.QWidget()
    
    return widget

def _set_policy_on_widget_list(widget_list: List[QtWidgets.QWidget], policy: QtWidgets.QSizePolicy):
    for widget in widget_list:
        widget.setSizePolicy(policy)

class ArgumentWidget(QtWidgets.QWidget):
    """
    Base class for other more complex widgets to inherit. An instance of `scrilla.gui.widgets.ArgumentWidget` embeds its child widgets in its own layout,  which is an instance of``PySide6.QtWidgetQVBoxLayout``. This class provides access to the following widgets:a ``PySide6.QtWidgets.QLabel`` for the title widget, an error message widget, a ``PySide6.QtWidgets.QPushButton`` for the calculate button widget, a ``PySide6.QtWidget.QLineEdit`` for ticker symbol input, and a variety of optional input widgets enabled by boolean flags in the `controls` constructor argument. 
    
    Constructor
    -----------
    1. **calculate_function**: ``str``
    2. **clear_function**: ``str``
    3. **controls:**: ``Dict[bool]``
    4. **widget_title**: ``str``
        *Optional*. Defaults to `Function Input`. Title affixed to the top of the widget.
    5. **button_msg**: ``str``
        *Optional*. Defaults to `Calculate`. Message written on `self.calculate_button`

    Attributes
    ----------
    1. **title**: ``PySide6.QtWidgets.QLabel``
    2. **required_title**: ``PySide6.QtWidgets.QLabel``
    3. **optional_title**: ``PySide6.QtWidgets.QLabel``
    4. **required_pane**: ``PySide6.QtWidgets.QLabel``
    5. **optional_pane**: ``PySide6.QtWidgets.QLabel``
    6. **message**: ``PySide6.QtWidgets.QLabel``
        Label attached to `self.symbol_input`.
    7. **error_message**: ``PySide6.QtWidget.QLabel``
        Message displayed if there is an error thrown during calculation.
    8. **calculate_button**: ``PySide6.QtWidget.QPushButton``
        Empty button for super classes to inherit, primed to fire events on return presses
    9. **clear_button**: ``PySide6.QtWidget.QPushButton``
        Empty button for super classes to inherit, primed to fire events on return presses.
    10. **symbol_widget**: ``PySide6.QtWidget.QLineEdit``
        Empty text area for super classes to inherit
    """
    # TODO: calculate and clear should be part of THIS constructor, doesn't make sense to have other widgets hook them up.
    def __init__(self, calculate_function: Callable, clear_function: Callable, controls: Dict[str, bool], layer):
        super().__init__()
        self.layer = layer
        self.controls = controls
        self.control_widgets = {}
        self.calculate_function = calculate_function
        self.clear_function = clear_function

        self._init_arg_widgets()
        self._arrange_arg_widgets()
        self._stage_arg_widgets()
    
    def _init_arg_widgets(self):
        """Creates child widgets"""
        self.title = _atomic_widget_factory(format='subtitle', title='Function Input')
        self.required_title = _atomic_widget_factory(format='label', title='Required Arguments')
        self.optional_title = _atomic_widget_factory(format='label', title='Optional Arguments')
        self.error_message = _atomic_widget_factory(format='error', title="Error Message Goes Here")
        self.calculate_button = _atomic_widget_factory(format='button', title='Calculate')
        self.clear_button = _atomic_widget_factory(format='button', title='Clear')
        self.symbol_hint = _atomic_widget_factory(format='text', title="Separate symbols with comma")

        self.symbol_widget = _composite_widget_factory(format='line-edit', title="Symbols :")
        for control in self.controls:
            if self.controls[control]:
                self.control_widgets[control] = _composite_widget_factory(definitions.ARG_DICT[control]['widget_type'], 
                                                                            f'{definitions.ARG_DICT[control]["name"]} :')
            else:
                self.control_widgets[control] = None

        self.required_pane = _layout_factory(format='vertical-box')
        self.required_pane.setObjectName(self.layer)
        self.optional_pane =_layout_factory(format='vertical-box')
        self.optional_pane.setObjectName(self.layer)
        self.setLayout(QtWidgets.QVBoxLayout())
    
    def _arrange_arg_widgets(self):
        """Arrange child widgets in their layouts and provides rendering hints"""
        self.symbol_hint.setAlignment(QtCore.Qt.AlignBottom)
        self.error_message.setAlignment(QtCore.Qt.AlignHCenter)
        
        _set_policy_on_widget_list([self.title, self.required_title, self.optional_title, self.calculate_button, 
                                    self.clear_button, self.symbol_hint, self.symbol_widget, self.required_pane, 
                                    self.optional_pane], 
                                    QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Maximum))
        _set_policy_on_widget_list([self.symbol_widget], 
                                    QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Maximum))

        self.required_pane.layout().addWidget(self.required_title)
        self.required_pane.layout().addWidget(self.symbol_hint)
        self.required_pane.layout().addWidget(self.symbol_widget)

        self.optional_pane.layout().addWidget(self.optional_title)
        for control_widget in self.control_widgets.values():
            if control_widget is not None:
                self.optional_pane.layout().addWidget(control_widget)

        self.layout().addWidget(self.title)
        self.layout().addWidget(self.error_message)
        self.layout().addWidget(self.required_pane)
        self.layout().addWidget(self.optional_pane)
        self.layout().addStretch()
        self.layout().addWidget(self.calculate_button)
        self.layout().addWidget(self.clear_button)

    def _stage_arg_widgets(self):
        """Prepares child widgets for display"""
        self.error_message.hide()
        self.calculate_button.setAutoDefault(True) # emits 'clicked' when return is pressed
        self.clear_button.setAutoDefault(True)
        self.clear_button.clicked.connect(self.clear_function)
        self.calculate_button.clicked.connect(self.calculate_function)
        self.get_symbol_input().setMaxLength(100)
        self.get_symbol_input().returnPressed.connect(self.calculate_function)

    def get_symbol_input(self):
        return self.symbol_widget.layout().itemAt(1).widget()

    def get_control_input(self, control_widget_key):
        if self.control_widgets[control_widget_key] is None:
            return None
        return self.control_widgets[control_widget_key].layout().itemAt(1).widget()


class TableWidget(QtWidgets.QWidget):
    """
    Base class for other more complex widgets to inherit. An instance of `scrilla.gui.widgets.TableWidget` embeds its child widgets in its own layout, which is an instance of``PySide6.QtWidgetQVBoxLayout``. This class provides access to the following widgets: a ``PySide6.QtWidgets.QLabel`` for the title widget, an error message widget and a ``PySide6.QtWidgets.QTableWidget`` for results display. 

    Parameters
    ----------
    1. **widget_title**: ``str``
        Title of the widget. Passed to the super class `scrilla.gui.widgets.SymbolWidget`.
    2. **button_msg**: ``str``
        Message written on the `self.calculate_button`. Passed to the super class `scrilla.gui.widgets.SymbolWidget`.

    Attributes
    ----------
    1. **title**: ``PySide6.QtWidget.QLabel``
    2.. **table**: ``PySide6.QtWidget.QTableWidget``

    """
    def __init__(self, widget_title: str ="Table Result"):
        super().__init__()
        self._init_table_widgets(widget_title)
        self._arrange_table_widgets()
        self._stage_table_widgets()

    def _init_table_widgets(self, widget_title):
        """Creates child widgets and their layouts"""
        self.title = QtWidgets.QLabel(widget_title)
        self.title.setObjectName('heading')

        self.table = QtWidgets.QTableWidget()
        self.table.setObjectName('table')

        self.setLayout(QtWidgets.QVBoxLayout())
    
    def _arrange_table_widgets(self):
        self.table.setHorizontalHeader(QtWidgets.QHeaderView(QtCore.Qt.Horizontal))
        self.table.horizontalHeader().setSectionResizeMode(QtWidgets.QHeaderView.Stretch)

        self.table.setVerticalHeader(QtWidgets.QHeaderView(QtCore.Qt.Vertical))
        # self.table.verticalHeader().setSectionResizeMode(QtWidgets.QHeaderView.Stretch)

        self.layout().addWidget(self.title)
        self.layout().addStretch()
        self.layout().addWidget(self.table, 1)

    def _stage_table_widgets(self):
        self.table.setSizeAdjustPolicy(QtWidgets.QAbstractScrollArea.AdjustToContents)
        self.table.hide()
    
    def show_table(self):
        self.table.resizeColumnsToContents()
        self.table.show()

class GraphWidget(QtWidgets.QWidget):
    """
    Base class for other more complex widgets to inherit. An instance of `scrilla.gui.widgets.GraphWidget` embeds its child widgets in its own layout, which is an instance of``PySide6.QtWidgetQVBoxLayout``. This class provides access to the following widgets: a ``PySide6.QtWidgets.QLabel`` for the title widget, an error message widget and a ``PySide6.QtWidgets.QTableWidget`` for results display. 

    The graph to be displayed should be stored in the `scrilla.settings.TEMP_DIR` directory before calling the `show_pixmap` method on this class. It will load the graph from file and convert it into a ``PySide6.QtGui.QPixMap``.

    Parameters
    ----------
    1. **tmp_graph_key**: ``str``
        The key of the file in the `scrilla.settings.TEMP_DIR` used to store the image of the graph.
    2. **widget_title**: ``str``
        Title applied to the label of the graph.

    Attributes
    ----------
    1. **tmp_graph_key**: ``str``
    2. **title**: ``PySide6.QtWidget.QLabel``
    3. **figure**: ``PySide6.QtWidget.QLabel``
    """
    def __init__(self, tmp_graph_key: str, widget_title: str ="Graph Results"):
        super().__init__()    
        self.tmp_graph_key = tmp_graph_key
        self._init_graph_widgets(widget_title)
        self._arrange_graph_widgets()
        self._stage_graph_widgets() 
    
    def _init_graph_widgets(self, widget_title):
        self.title = QtWidgets.QLabel(widget_title)
        self.title.setObjectName('heading')
        self.figure = QtWidgets.QLabel()
        self.setLayout(QtWidgets.QVBoxLayout())


    def _arrange_graph_widgets(self):
        self.title.setSizePolicy(QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding,
                                                            QtWidgets.QSizePolicy.Minimum))
        self.figure.setSizePolicy(QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding,
                                                            QtWidgets.QSizePolicy.Expanding))

        self.title.setAlignment(QtCore.Qt.AlignLeft)
        self.figure.setAlignment(QtCore.Qt.AlignCenter)

        self.layout().addWidget(self.title)
        self.layout().addWidget(self.figure)
    
    def _stage_graph_widgets(self):
        self.figure.hide()

    def set_pixmap(self):
        self.figure.setPixmap(utilities.generate_pixmap_from_temp(self.width(), self.height(), self.tmp_graph_key))
        self.figure.show()

class CompositeWidget(QtWidgets.QWidget):
    """
    Constructor
    -----------
    1. **tmp_graph_key**: ``str``
    2. **widget_title**: ``str``
        Title of the widget. Passed to the super class `scrilla.gui.widgets.SymbolWidget`.
    3. **table_title**: ``str``
    4. **graph_title**: ``str``

    Attributes
    ----------
    1. **title**: ``PySide6.QtWidgets.QLabel``
    2. **table_widget**: ``scrilla.gui.widgets.TableWidget``
    3. **graph_widget**: ``scrilla.gui.widgets.GraphWidget``
    4. **tab_widget**: ``PySide6.QtWidget.QWidget``

    """
    def __init__(self, tmp_graph_key: str, layer: str, widget_title: str="Results", 
                    table_title: Union[str, None]=None, 
                    graph_title: Union[str,None]=None):
        super().__init__()
        self.setObjectName(layer)
        self._init_composite_widgets(widget_title=widget_title,
                                        tmp_graph_key=tmp_graph_key)
        self._arrange_composite_widgets(graph_title=graph_title, 
                                        table_title=table_title)

    def _init_composite_widgets(self, widget_title, tmp_graph_key):
        """Creates child widgets and their layouts"""
        self.title = QtWidgets.QLabel(widget_title)
        self.title.setObjectName('subtitle')

        self.table_widget = TableWidget()
        self.graph_widget = GraphWidget(tmp_graph_key)

        self.tab_widget = QtWidgets.QTabWidget()

        self.setLayout(QtWidgets.QVBoxLayout())
        
    def _arrange_composite_widgets(self, graph_title, table_title):
        self.title.setSizePolicy(QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding,
                                                            QtWidgets.QSizePolicy.Minimum))
        self.title.setAlignment(QtCore.Qt.AlignLeft)

        self.tab_widget.addTab(self.table_widget, table_title)
        self.tab_widget.addTab(self.graph_widget, graph_title)
        
        self.layout().addWidget(self.title)
        self.layout().addWidget(self.tab_widget)



import os, datetime
from typing import List, Union
import numpy, matplotlib
from PIL import Image
from PIL.Image import Image as img

from matplotlib.figure import Figure
from matplotlib import dates as mdates

from scrilla import settings
from scrilla.static import formats, keys
from scrilla.analysis.objects.portfolio import Portfolio
from scrilla.analysis.objects.cashflow import Cashflow
from scrilla.errors import InputValidationError
from scrilla.util import dater

APP_ENV=os.environ.setdefault('APP_ENV', 'local')

if APP_ENV == 'local':
    from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
    matplotlib.use("Qt5Agg")
elif APP_ENV == 'container':
    from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas
    matplotlib.use("agg")

def _show_or_save(canvas: FigureCanvas, show: bool = True, savefile: str = None) -> Union[FigureCanvas, None]:
    if savefile is not None:
        canvas.print_jpeg(filename_or_obj=savefile)

    s, (width, height) = canvas.print_to_buffer()
    im = Image.frombytes("RGBA", (width, height), s)

    if show:
        im.show()
        return None
    canvas.draw()
    return canvas

def plot_qq_series(ticker: str, sample: list, show: bool=True, savefile: str=None) -> Union[FigureCanvas, None]:
    title = f'{ticker} Return Q-Q Plot'
    normal_series = []
    sample_series = []
    for point in sample:
        normal_series.append(point[0])
        sample_series.append(point[1])

    canvas = FigureCanvas(Figure())
    axes = canvas.figure.subplots()

    axes.plot(normal_series, sample_series, linestyle="None", marker=".", markersize=10.0)
    axes.plot(normal_series, normal_series)
    axes.grid()
    axes.set_xlabel('Normal Percentiles')
    axes.set_ylabel('Sample Percentiles')
    axes.set_title(title)

    return _show_or_save(canvas=canvas, show=show, savefile=savefile)

def plot_correlation_series(tickers: list, series: dict, show: bool=True, savefile: str=None) -> Union[FigureCanvas, None]:
    start, end = list(series.keys())[-1], list(series.keys())[0]
    title = f'({tickers[0]}, {tickers[1]}) correlation time series'
    subtitle = f'{start} to {end}, rolling {settings.DEFAULT_ANALYSIS_PERIOD}-day estimate'

    canvas = FigureCanvas(Figure())
    figure = canvas.figure
    axes = figure.subplots()

    locator = mdates.AutoDateLocator()
    formatter = mdates.AutoDateFormatter(locator)

    correl_history, dates= [], []
    for date in series:
        dates.append(dater.parse_date_string(date))
        correl_history.append(series[date])

    axes.plot(dates, correl_history)
    axes.grid()
    axes.xaxis.set_major_locator(locator)
    axes.xaxis.set_major_formatter(formatter)
    axes.set_ylabel('Correlation')
    axes.set_xlabel('Dates')
    axes.set_title(subtitle, fontsize=12)
    figure.suptitle(title, fontsize=18)
    figure.autofmt_xdate()

    return _show_or_save(canvas=canvas, show=show, savefile=savefile)

def plot_frontier(portfolio: Portfolio, frontier: list, show: bool=True, savefile: str=None) -> Union[FigureCanvas, None]:
    title = " ("
    for i, item in enumerate(portfolio.tickers):
        if i != (len(portfolio.tickers) - 1):
            title += item + ", "
        else:
            title += item + ") Efficient Frontier"
    
    return_profile, risk_profile = [], []
    for allocation in frontier:
        return_profile.append(portfolio.return_function(allocation))
        risk_profile.append(portfolio.volatility_function(allocation))

    canvas = FigureCanvas(Figure())
    axes = canvas.figure.subplots()

    axes.plot(risk_profile, return_profile, linestyle='dashed')
    axes.grid()
    axes.set_xlabel('Volatility')
    axes.set_ylabel('Return')
    axes.set_title(title)

    return _show_or_save(canvas=canvas, show=show, savefile=savefile)

def plot_yield_curve(yield_curve: dict, show: bool=True, savefile: str=None) -> Union[FigureCanvas, None]:
    title = f'US Treasury Yield Curve On {list(yield_curve.keys())[0]}'

    canvas = FigureCanvas(Figure())
    axes = canvas.figure.subplots()

    maturities, rates = [], yield_curve[list(yield_curve.keys())[0]]
    yield_map = keys.keys['SERVICES']['STATISTICS']['QUANDL']['MAP']['YIELD_CURVE']
    for i, rate in enumerate(rates):
        maturities.append(yield_map[keys.keys['YIELD_CURVE'][i]])

    axes.plot(maturities, rates, linestyle="dashed", marker= ".", markersize=10.0)
    axes.grid()
    axes.set_xlabel('Maturity')
    axes.set_ylabel('Annual Yield')
    axes.set_title(title)

    return _show_or_save(canvas=canvas, show=show, savefile=savefile)

def plot_profiles(symbols: list, profiles: dict, show: bool=True, savefile: str=None, subtitle: str=None) -> Union[FigureCanvas, None]:
    canvas = FigureCanvas(Figure())

    no_symbols = len(symbols)
    axes = canvas.figure.subplots()

    title ="("
    for symbol in symbols:
        if symbols.index(symbol) != (len(symbols)-1):
            title += symbol +", "
        else:
            title += symbol +') Risk-Return Profile'
            if subtitle is not None:
                title += "\n" + subtitle

    return_profile, risk_profile = [], []
    for profile in profiles:
        return_profile.append(profiles[profile]['annual_return'])
        risk_profile.append(profiles[profile]['annual_volatility'])

    axes.plot(risk_profile, return_profile, linestyle='None', marker= ".", markersize=10.0)
    axes.grid()
    axes.set_xlabel('Volatility')
    axes.set_ylabel('Return')
    axes.set_title(title)

    for i in range(no_symbols):
        axes.annotate(symbols[i], (risk_profile[i], return_profile[i]))

    return _show_or_save(canvas=canvas, show=show, savefile=savefile)

    # TODO: figure out date formatting for x-axis

def plot_moving_averages(symbols:List[str], averages_output: List[List[float]], periods: List[int], show: bool=True, savefile: str=None) -> Union[FigureCanvas, None]:
    averages, dates = averages_output
    canvas = FigureCanvas(Figure())
    axes = canvas.figure.subplots()
    ma1_label, ma2_label, ma3_label = f'MA({periods[0]})', f'MA({periods[1]})', f'MA({periods[2]})'

    if dates is None:
        ma1s, ma2s, ma3s = [], [], []
        for i in range(len(symbols)):
            ma1s.append(averages[i][0])
            ma2s.append(averages[i][1])
            ma3s.append(averages[i][2])
    
        width = formats.formats['BAR_WIDTH']
        x = numpy.arange(len(symbols))

        axes.bar(x + width, ma1s, width, label=ma1_label)
        axes.bar(x, ma2s, width, label=ma2_label)
        axes.bar(x - width, ma3s, width, label=ma3_label)

        axes.set_ylabel('Annual Logarithmic Return')
        axes.set_title('Moving Averages of Annualized Daily Return Grouped By Equity')
        axes.set_xticks(x)
        axes.set_xticklabels(symbols)
        axes.legend()

    else:
        
        # TODO: generate different locators based on length of period
        x = [datetime.datetime.strptime(dater.date_to_string(date), '%Y-%m-%d').toordinal() for date in dates]
        date_locator = matplotlib.dates.WeekdayLocator(byweekday=(matplotlib.dates.WE))
        date_format = matplotlib.dates.DateFormatter('%m-%d')
        
        for i, item in enumerate(symbols):
            ma1s, ma2s, ma3s = [], [], []
            ma1_label, ma2_label, ma3_label = f'{item}_{ma1_label}', f'{item}_{ma2_label}', f'{item}_{ma3_label}'
            for j in range(len(dates)):
                MA_1 = averages[i][0][j]
                ma1s.append(MA_1)
                                  
                MA_2 = averages[i][1][j]
                ma2s.append(MA_2)
         
                MA_3 = averages[i][2][j]
                ma3s.append(MA_3)
            
            start_date, end_date = dates[0], dates[-1] 
            title_str = f'Moving Averages of Annualized Return From {start_date} to {end_date}'

            axes.plot(x, ma1s, linestyle="solid", color="darkgreen", label=ma1_label)
            axes.plot(x, ma2s, linestyle="dotted", color="gold", label=ma2_label)
            axes.plot(x, ma3s, linestyle="dashdot", color="orangered", label=ma3_label)

        axes.set_title(title_str)
        axes.set_ylabel('Annualized Logarthmic Return')
        axes.set_xlabel('Dates')
        axes.xaxis.set_major_locator(date_locator)
        axes.xaxis.set_major_formatter(date_format)
        
        axes.legend()

    return _show_or_save(canvas=canvas, show=show, savefile=savefile)

def plot_cashflow(ticker: str, cashflow: Cashflow, show: bool=True, savefile: str=None) -> Union[FigureCanvas, None]:
    if not cashflow.beta or not cashflow.alpha or len(cashflow.sample) < 3:
        raise InputValidationError("Cashflow model does not contain enough information to be plotted")
    
    canvas = FigureCanvas(Figure())
    figure = canvas.figure
    axes = figure.subplots()

    date_format = mdates.DateFormatter('%m-%d')

    sup_title_str = f'{ticker} Dividend Linear Regression Model'
    title_str = f'NPV(dividends | discount = {round(cashflow.discount_rate,4)}) = $ {round(cashflow.calculate_net_present_value(), 2)}'

    dividend_history, ordinal_x, dates= [], [], []
    for date in cashflow.sample:
        dates.append(date)
        ordinal_x.append(datetime.datetime.strptime(date, '%Y-%m-%d').toordinal())
        dividend_history.append(cashflow.sample[date])

    dates.reverse()
    
    model_map = list(map(lambda x: cashflow.alpha + cashflow.beta*x, cashflow.time_series))
    
    axes.scatter(ordinal_x, dividend_history, marker=".")
    axes.plot(ordinal_x, model_map)
    axes.grid()
    # axes.xaxis.set_major_formatter(date_format)
    axes.set_xticklabels(dates)
    axes.set_ylabel('Dividend Amount')
    axes.set_xlabel('Payment Date')
    axes.set_title(title_str, fontsize=12)
    figure.suptitle(sup_title_str, fontsize=18)
    figure.autofmt_xdate()

    return _show_or_save(canvas=canvas, show=show, savefile=savefile)



from scrilla.static.constants import constants
from scrilla.static import keys
from scrilla.util import dater

def get_trading_period(asset_type: str) -> float:
    """
    Returns the value of one trading day measured in years of the asset_type passed in as an argument.

    Parameters
    ----------
    1. **asset_type**: ``str``
    
    A string that represents a type of tradeable asset. Types are statically accessible through the ` settings` variables: ASSET_EQUITY and ASSET_CRYPTO.
    """
    if asset_type == keys.keys['ASSETS']['CRYPTO']:
        return constants['ONE_TRADING_DAY']['CRYPTO']
    return constants['ONE_TRADING_DAY']['EQUITY']

def format_profiles(profiles: dict):
    profiles_format = []
    for key, value in profiles.items():
        holding = value
        holding['ticker'] = key
        profiles_format.append(holding)
    return profiles_format

def format_allocation(allocation, portfolio, investment=None):
    allocation_format = []

    if investment is not None:
        shares = portfolio.calculate_approximate_shares(x=allocation, total=investment)
        total = portfolio.calculate_actual_total(x=allocation, total=investment)

    annual_volatility = portfolio.volatility_function(x=allocation) 
    annual_return = portfolio.return_function(x=allocation)

    for j, item in enumerate(portfolio.tickers):
        holding = {}
        holding['ticker'] = item
        holding['allocation'] = round(allocation[j], constants['ACCURACY'])
        if investment is not None:
            holding['shares'] = float(shares[j])
        holding['annual_return'] = round(portfolio.mean_return[j], constants['ACCURACY']) 
        holding['annual_volatility'] = round(portfolio.sample_vol[j], constants['ACCURACY'])
        allocation_format.append(holding)

    json_format = {}
    json_format['holdings'] = allocation_format

    if investment is not None:
        json_format['total'] = float(total)
        
    json_format['portfolio_return'] = annual_return
    json_format['portfolio_volatility'] = annual_volatility
    
    return json_format

def format_frontier(portfolio, frontier, investment=None):
    json_format = []
    for i, item in enumerate(frontier):
        json_format.append(format_allocation(allocation=item, portfolio=portfolio, 
                                                            investment=investment))
    return json_format

def format_correlation_matrix(tickers, correlation_matrix):
    response = []
    for i, item in enumerate(tickers):
        # correlation_matrix[i][i]
        for j in range(i+1, len(tickers)):
            subresponse = {}
            subresponse[f'{item}_{tickers[j]}_correlation'] = correlation_matrix[j][i]
            response.append(subresponse)
    return response

# GUI FORMATTING
def format_allocation_profile_title(allocation, portfolio) -> str:
    port_return, port_volatility = portfolio.return_function(allocation), portfolio.volatility_function(allocation)
    formatted_result = "("+str(100*port_return)[:5]+"%, " + str(100*port_volatility)[:5]+"%)"
    formatted_result_title = "("
    for symbol in portfolio.tickers:
        if portfolio.tickers.index(symbol) != (len(portfolio.tickers) - 1):
            formatted_result_title += symbol+", "
        else:
            formatted_result_title += symbol + ") Portfolio Return-Risk Profile"
    whole_thing = formatted_result_title +" = "+formatted_result
    return whole_thing

# TODO: refactor this, along with scrilla.analysis.models.geometric.statistics.calculate_moving_averages
def format_moving_averages(tickers, averages_output):
    these_moving_averages, dates = averages_output

    response = {}
    for i, item in enumerate(tickers):
        ticker_str=f'{item}'
        MA_1_str, MA_2_str, MA_3_str = f'{ticker_str}_MA_1', f'{ticker_str}_MA_2', f'{ticker_str}_MA_3'    

        subresponse = {}
        if dates is None:
            subresponse[MA_1_str] = these_moving_averages[i][0]
            subresponse[MA_2_str] = these_moving_averages[i][1]
            subresponse[MA_3_str] = these_moving_averages[i][2]

        else:
            subsubresponse_1, subsubresponse_2, subsubresponse_3 = {}, {}, {}
    
            for j, this_item in enumerate(dates):
                date_str=dater.date_to_string(this_item)
                subsubresponse_1[date_str] = these_moving_averages[i][0][j]
                subsubresponse_2[date_str] = these_moving_averages[i][1][j]
                subsubresponse_3[date_str] = these_moving_averages[i][2][j]

            subresponse[MA_1_str] = subsubresponse_1
            subresponse[MA_2_str] = subsubresponse_2
            subresponse[MA_3_str] = subsubresponse_3

        response[ticker_str] = subresponse
    
    return response



# This file is part of scrilla: https://github.com/chinchalinchin/scrilla.

# scrilla is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3
# as published by the Free Software Foundation.

# scrilla is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with scrilla.  If not, see <https://www.gnu.org/licenses/>
# or <https://github.com/chinchalinchin/scrilla/blob/develop/main/LICENSE>.
"""
Application-wide configuration settings. 
"""
import os, json

import scrilla.util.outputter as outputter

class APIKeyError(Exception):
    pass

APP_DIR = os.path.dirname(os.path.abspath(__file__))
"""Folder containing the root module of the project"""

PROJECT_DIR = os.path.dirname(APP_DIR)
"""Folder containing the project source"""

APP_NAME="scrilla"
"""Name of the application"""

APP_ENV = os.environ.setdefault('APP_ENV', 'local')
"""Application execution environment; Configured by environment variable of the same name, **APP_ENV**."""

LOG_LEVEL = str(os.environ.setdefault("LOG_LEVEL", "info")).lower()
"""Application log level output; Configured by environment of the same name, **LOG_LEVEL**."""
logger = outputter.Logger('settings', LOG_LEVEL)

# TODO: save formatting only supports JSON currently. Future file extensions: csv and txt.
FILE_EXT = os.environ.setdefault("FILE_EXT", "json")
"""Extension used to save files; Configured by environment variable of the same name, **FILE_EXT**"""

CACHE_DIR = os.path.join(APP_DIR, 'data', 'cache')
"""Directory containing cached prices, statistics and calculations"""
CACHE_SQLITE_FILE = os.environ.setdefault('SQLITE_FILE', os.path.join(CACHE_DIR, 'scrilla.db'))
"""Location of the SQLite database flat file; Configured by environment variable **SQLITE_FILE***"""

TEMP_DIR = os.path.join(APP_DIR, 'data', 'tmp')

STATIC_DIR = os.path.join(APP_DIR, 'data', 'static')
"""Directory containg static data, such as ticker symbols, statistic symbols, etc."""

STATIC_TICKERS_FILE = os.path.join(STATIC_DIR, f'tickers.{FILE_EXT}')
"""Location of file used to store equity ticker symbols"""
STATIC_ECON_FILE = os.path.join(STATIC_DIR, f'economics.{FILE_EXT}')
"""Location of file used to store statistic symbols"""
STATIC_CRYPTO_FILE = os.path.join(STATIC_DIR, f'crypto.{FILE_EXT}')
"""Location of file used to store crypto ticker symbols"""

COMMON_DIR=os.path.join(APP_DIR, 'data', 'common')
"""Directory used to store common files, such as API keys, watchlist, etc."""
COMMON_WATCHLIST_FILE=os.path.join(COMMON_DIR, f'watchlist.{FILE_EXT}')
"""Location of file used to store watchlisted ticker symbols"""

GUI_STYLESHEET_FILE=os.path.join(APP_DIR, 'gui', 'styles', 'app.qss')
"""Location of the stylesheet applied to the GUI"""

GUI_THEME_FILE=os.path.join(APP_DIR, 'gui', 'styles', 'themes.json')

GUI_DARK_MODE = os.environ.setdefault('DARK_MODE', 'true').lower() == 'true'

## GUI CONFIGURATION
try:
    GUI_WIDTH = int(os.environ.setdefault('GUI_WIDTH', '1024'))
    """Width of main Graphical User Interface window; Configured by environment variable of same name, **GUI_WIDTH**"""
except (ValueError, TypeError) as ParseError: 
    logger.debug('Failed to parse GUI_WIDTH from environment. Setting to default value of 1024.')
    GUI_WIDTH = 1024
    os.environ['GUI_WIDTH'] = '1024'

try:
    GUI_HEIGHT = int(os.environ.setdefault('GUI_HEIGHT', '768'))
    """Height of main Graphical User Interface window; Configured by environment variable of same name, **GUI_HEIGHT**."""
except (ValueError, TypeError) as ParseError:
    logger.debug('Failed to parse GUI_HEIGHT from enviroment. Setting to default value of 768.')
    GUI_HEIGHT = 768
    os.environ['GUI_HEIGHT'] = '768'

## FINANCIAL ALGORITHM CONFIGURATION
try:
    FRONTIER_STEPS = int(os.environ.setdefault('FRONTIER_STEPS', '5'))
    """Number of data points used to trace out the efficient frontier; Configured by environment variable of same name, **FRONTIER_STEPS** """
except (ValueError, TypeError) as ParseError:
    logger.debug('Failed to parse FRONTIER_STEPS from enviroment. Setting to default value of 5.')
    FRONTIER_STEPS = 5
    os.environ['FRONTIER_STEPS'] = '5'

try:
    MA_1_PERIOD = int(os.environ.setdefault('MA_1', '20'))
    """Number of data points in first moving average period; Configured by environment variable, **MA_1**"""
except (ValueError, TypeError) as ParseError: 
    logger.debug('Failed to parse MA_1 from environment. Setting to default value of 20.')
    MA_1_PERIOD = 20
    os.environ['MA_1'] = '20'

try:
    MA_2_PERIOD = int(os.environ.setdefault('MA_2', '60'))
    """Number of data points in second moving average period; Configured by environment variable, **MA_2**"""
except (ValueError, TypeError) as ParseError: 
    logger.debug('Failed to parse MA_2 from environment. Setting to default value of 60.')
    MA_2_PERIOD = 60
    os.environ['MA_2'] = '60'

try:
    MA_3_PERIOD = int(os.environ.setdefault('MA_3', '100'))
    """Number of data points in third moving average period; Configured by environment variable, **MA_3***"""
except (ValueError, TypeError) as ParseError: 
    logger.debug('Failed to parse MA_3 from environment. Setting to default value of 100.')
    MA_3_PERIOD = 100
    os.environ['MA_3'] = '100'

try:
    ITO_STEPS = int(os.environ.setdefault('ITO_STEPS', '10000'))
    """Number of iterations used to approximate an Ito integral; Configured by environment variable of same name, **ITO_STEPS**"""
except (ValueError, TypeError) as ParseError:
    logger.debug('Failed to parse ITO_STEPS from environment. Setting to default of 10000.')
    ITO_STEPS = 10000
    os.environ['ITO_STEPS'] = '10000'

try:
    DEFAULT_ANALYSIS_PERIOD=int(os.environ.setdefault('DEFAULT_ANALYSIS_PERIOD', '100'))
    """Number of days used in a historical sample, if no date range is specified; Configured by environment variable of same name, **DEFAULT_ANALYSIS_PERIOD**"""
except (ValueError, TypeError) as ParseError:
    logger.debug('Failed to parse DEFAULT_ANALYSIS_PERIOD from environment. Setting to default of 100.')
    DEFAULT_ANALYSIS_PERIOD=100
    os.environ['DEFAULT_ANALYSIS_PERIOD']=100

RISK_FREE_RATE=os.environ.setdefault("RISK_FREE", "TEN_YEAR").strip("\"")
"""Maturity of the US Treasury Yield Curve used as a proxy for the risk free rate"""

MARKET_PROXY=os.environ.setdefault('MARKET_PROXY', 'SPY')
"""Ticker symbol used as a proxy for the overall market rate of return"""

ANALYSIS_MODE=os.environ.setdefault('ANALYSIS_MODE', 'geometric')
"""Determines the asset price process and thus, the underlying probability distribution of an asset's return"""

ESTIMATION_METHOD=os.environ.setdefault('DEFAULT_ESTIMATION_METHOD', 'moments')
"""Determines the default estimation method using in statistical estimations"""

## SERVICE CONFIGURATION
### PRICE_MANAGER CONFIGRUATION
PRICE_MANAGER = os.environ.setdefault('PRICE_MANAGER', 'alpha_vantage')
"""Determines the service used to retrieve price data"""

#### ALPHAVANTAGE CONFIGURATION
if PRICE_MANAGER == 'alpha_vantage':
    AV_URL = os.environ.setdefault('ALPHA_VANTAGE_URL', 'https://www.alphavantage.co/query').strip("\"").strip("'")
    AV_CRYPTO_LIST=os.environ.setdefault('ALPHA_VANTAGE_CRYPTO_META_URL', 'https://www.alphavantage.co/digital_currency_list/')

    AV_KEY = os.environ.setdefault('ALPHA_VANTAGE_KEY', '')
    if not AV_KEY:
        keystore = os.path.join(COMMON_DIR, f'ALPHA_VANTAGE_KEY.{FILE_EXT}')
        if os.path.isfile(keystore):
            with open(keystore, 'r') as infile:
                if FILE_EXT == "json":
                    AV_KEY = json.load(infile)['ALPHA_VANTAGE_KEY']
                    os.environ['ALPHA_VANTAGE_KEY'] = str(AV_KEY)

    if not AV_KEY:
        raise APIKeyError('Alpha Vantage API Key not found. Either set ALPHA_VANTAGE_KEY environment variable or use "-store" CLI function to save key.')

### STAT_MANAGER CONFIGURATION
STAT_MANAGER = os.environ.setdefault('STAT_MANAGER', 'quandl')
"""Determines the service used to retrieve statistics data"""

#### QUANDL CONFIGURAITON
if STAT_MANAGER == "quandl":
    Q_URL = os.environ.setdefault('QUANDL_URL', 'https://www.quandl.com/api/v3/datasets').strip("\"").strip("'")
    Q_META_URL = os.environ.setdefault('QUANDL_META_URL' ,'https://www.quandl.com/api/v3/databases')
    Q_KEY = os.environ.setdefault('QUANDL_KEY', '')

    if not Q_KEY:
        keystore = os.path.join(COMMON_DIR, f'QUANDL_KEY.{FILE_EXT}')
        if os.path.isfile(keystore):
            with open(keystore, 'r') as infile:
                if FILE_EXT == "json":
                    Q_KEY = json.load(infile)['QUANDL_KEY']
                    os.environ['QUANDL_KEY'] = str(Q_KEY)

    if not Q_KEY:
        raise APIKeyError('Quandl API Key not found. Either set QUANDL_KEY environment variable or use "-store" CLI function to save key.')

### DIVIDEND_MANAGER CONFIGURATION
DIV_MANAGER=os.environ.setdefault("DIV_MANAGER", 'iex')
"""Determines the service used to retrieve dividends data"""

if DIV_MANAGER == "iex":
    IEX_URL = os.environ.setdefault("IEX_URL", 'https://cloud.iexapis.com/stable/stock')
    
    IEX_KEY = os.environ.setdefault("IEX_KEY", '')
    if not IEX_KEY:
        keystore = os.path.join(COMMON_DIR, f'IEX_KEY.{FILE_EXT}')
        if os.path.isfile(keystore):
            with open(keystore, 'r') as infile:
                if FILE_EXT == "json":
                    IEX_KEY = json.load(infile)['IEX_KEY']
                    os.environ['IEX_KEY'] = str(IEX_KEY)

    if not IEX_KEY:
        raise APIKeyError('IEX API Key cannot be found. Either set IEX_KEY environment variable or use "-store" CLI function to save key.')



from PySide6.QtGui import QPixmap
from PySide6.QtCore import Qt

from scrilla import settings

def calculate_image_width(width) -> float:
    return 9*width/10

def calculate_image_height(height) -> float:
    return 9*height/10

def generate_pixmap_from_temp(width, height, ext) -> QPixmap:
    pixmap = QPixmap(f'{settings.TEMP_DIR}/{ext}')
    pixmap = pixmap.scaled(calculate_image_width(width), calculate_image_height(height), aspectMode=Qt.KeepAspectRatio)
    return pixmap

def get_next_layer(layer):
    if layer == "root":
        return "child"
    if layer == "child":
        return "grand-child"
    else:
        return f'great-{layer}'



import argparse
from math import trunc
from typing import List, Tuple

from scrilla.static import definitions, constants

def truncate(number: float, digits: int) -> float:
    stepper = 10.0 ** digits
    return trunc(stepper * number) / stepper

def strip_string_array(array: List[str]) -> List[str]:
    new_array = []
    for string in array:
        new_array.append(string.strip())
    return new_array

def split_and_strip(string: str, upper: bool = True, delimiter=",")-> List[str]:
    if upper:
        return strip_string_array(string.upper().split(delimiter))
    return strip_string_array(string.lower().split(delimiter))

def round_array(array: List[float], decimals: int) -> List[float]:
    cutoff = (1/10**decimals)
    return [0 if element < cutoff else truncate(element, decimals) for element in array]

def intersect_dict_keys(dict1: dict, dict2: dict) -> Tuple[dict, dict]:
    intersection = [x for x in dict1.keys() if x in dict2.keys()]

    new_dict1, new_dict2 = {}, {}
    for key, value in dict1.items():
        if key in intersection:
            new_dict1[key] = value
    for key, value in dict2.items():
        if key in intersection:
            new_dict2[key] = value
    
    return new_dict1, new_dict2

def format_float_number(decimal: float):
    return str(decimal)[:constants.constants['SIG_FIGS']]

def format_float_percent(decimal: float):
    return str(100*float(format_float_number(decimal)))[:constants.constants['SIG_FIGS']]+"%"

def format_dict_percent(this_dict: dict, which_key: str) -> dict:
    buffer_dict = this_dict.copy()
    buffer_dict[which_key] = format_float_percent(this_dict[which_key])
    return buffer_dict

def format_dict_number(this_dict: dict, which_key: str) -> dict:
    buffer_dict = this_dict.copy()
    buffer_dict[which_key] = format_float_number(this_dict[which_key])
    return buffer_dict

################################################
##### PARSING FUNCTIONS

### CLI PARSING
def format_args(args, default_estimation_method) -> argparse.Namespace:
    parser = argparse.ArgumentParser()

    choices = []
    for func in definitions.FUNC_DICT:
        choices.append(definitions.FUNC_DICT[func]['values'][0])
        choices.append(definitions.FUNC_DICT[func]['values'][1])

    parser.add_argument('function_arg', choices=choices)

    groups = [parser.add_mutually_exclusive_group() for arg_group in definitions.ARG_META['groups'] ]

    for arg in definitions.ARG_DICT:
        if definitions.ARG_DICT[arg]['format'] not in ('group', bool):
            parser.add_argument(definitions.ARG_DICT[arg]['values'][0], 
                                definitions.ARG_DICT[arg]['values'][1], 
                                definitions.ARG_DICT[arg]['values'][2], 
                                definitions.ARG_DICT[arg]['values'][3],
                                default=None, 
                                type=definitions.ARG_DICT[arg]['format'],
                                dest=arg)
        elif definitions.ARG_DICT[arg]['format'] == 'group':
            group_index = definitions.ARG_META['groups'].index(definitions.ARG_DICT[arg]['group'])
            groups[group_index].add_argument(definitions.ARG_DICT[arg]['values'][0],
                                            definitions.ARG_DICT[arg]['values'][1],
                                            definitions.ARG_DICT[arg]['values'][2],
                                            definitions.ARG_DICT[arg]['values'][3],
                                            action='store_const',
                                            dest=definitions.ARG_DICT[arg]['group'],
                                            const=arg)
        # NOTE: 'format' == group AND 'format' == bool => Empty Set, so only other alternative is
        # 'format' == bool
        else:
            parser.add_argument(definitions.ARG_DICT[arg]['values'][0], 
                                definitions.ARG_DICT[arg]['values'][1], 
                                definitions.ARG_DICT[arg]['values'][2], 
                                definitions.ARG_DICT[arg]['values'][3],
                                action='store_true',
                                dest=arg)

    parser.set_defaults(estimation_method=default_estimation_method)
    parser.add_argument('tickers', nargs='*', type=str)
    return vars(parser.parse_args(args))

def get_first_json_key(this_json: dict) -> str:
    return list(this_json.keys())[0]

def replace_troublesome_chars(msg: str) -> str:
    return msg.replace('\u2265','').replace('\u0142', '')



# This file is part of scrilla: https://github.com/chinchalinchin/scrilla.

# scrilla is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3
# as published by the Free Software Foundation.

# scrilla is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with scrilla.  If not, see <https://www.gnu.org/licenses/>
# or <https://github.com/chinchalinchin/scrilla/blob/develop/main/LICENSE>.

from PySide6 import QtGui, QtCore, QtWidgets


from scrilla import settings
from scrilla import static
from scrilla.static import keys
# TODO: conditional import based on ANALYSIS_MODE
from scrilla.analysis import markets, optimizer
from scrilla.analysis.models.geometric import statistics
from scrilla.analysis.objects.portfolio import Portfolio

from scrilla.util import outputter, helper, plotter

from scrilla.gui import formats, utilities
from scrilla.gui.widgets.components import ArgumentWidget, CompositeWidget, \
                                            GraphWidget, TableWidget, generate_control_skeleton

logger = outputter.Logger('gui.functions', settings.LOG_LEVEL)

class RiskReturnWidget(QtWidgets.QWidget):
    def __init__(self, layer):
        super().__init__()
        self.setObjectName(layer)
        self._init_profile_widgets()
        self._arrange_profile_widgets()

    @staticmethod
    def _configure_control_skeleton():
        controls = generate_control_skeleton()
        controls['start_date'] = True
        controls['end_date'] = True
        return controls

    def _init_profile_widgets(self):
        self.composite_widget = CompositeWidget(keys.keys['GUI']['TEMP']['PROFILE'], 
                                                    widget_title="Risk Analysis",
                                                    table_title="CAPM Risk Profile",
                                                    graph_title="Risk-Return Plane",
                                                    layer = utilities.get_next_layer(self.objectName()))
        self.arg_widget = ArgumentWidget(calculate_function=self.calculate,
                                            clear_function=self.clear,
                                            controls= RiskReturnWidget._configure_control_skeleton(),
                                            layer = utilities.get_next_layer(self.objectName()))
        self.setLayout(QtWidgets.QHBoxLayout())

    def _arrange_profile_widgets(self):
        self.arg_widget.setSizePolicy(QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Minimum,
                                                            QtWidgets.QSizePolicy.Expanding))
        self.layout().addWidget(self.composite_widget)
        self.layout().addWidget(self.arg_widget)

    @QtCore.Slot()
    def calculate(self):
        if self.composite_widget.graph_widget.figure.isVisible():
            self.composite_widget.graph_widget.figure.hide()

        symbols = helper.split_and_strip(self.arg_widget.get_symbol_input().text())

        self.composite_widget.table_widget.table.setRowCount(len(symbols))
        self.composite_widget.table_widget.table.setColumnCount(5)
        self.composite_widget.table_widget.table.setHorizontalHeaderLabels(['Return','Volatility', 'Sharpe', 'Beta', 'Equity Cost'])
        self.composite_widget.table_widget.table.setVerticalHeaderLabels(symbols)

        profiles = {}
        for i, symbol in enumerate(symbols):
            profiles[symbol] = statistics.calculate_risk_return(symbol)
            profiles[symbol][keys.keys['APP']['PROFILE']['SHARPE']] = markets.sharpe_ratio(symbol)
            profiles[symbol][keys.keys['APP']['PROFILE']['BETA']] = markets.market_beta(symbol)
            profiles[symbol][keys.keys['APP']['PROFILE']['EQUITY']] = markets.cost_of_equity(symbol)

            formatted_profile = formats.format_profile(profiles[symbol])

            for j, statistic in enumerate(formatted_profile.keys()):
                table_item = QtWidgets.QTableWidgetItem(formatted_profile[statistic])
                table_item.setTextAlignment(QtCore.Qt.AlignHCenter)
                self.composite_widget.table_widget.table.setItem(i, j, table_item)

        plotter.plot_profiles(symbols=symbols, profiles=profiles, show=False,
                                        savefile=f'{settings.TEMP_DIR}/{keys.keys["GUI"]["TEMP"]["PROFILE"]}')

        self.composite_widget.graph_widget.set_pixmap()
        self.composite_widget.table_widget.show_table()

    def resizeEvent(self, event: QtGui.QResizeEvent) -> None:
        if self.composite_widget.graph_widget.figure.isVisible():
            self.composite_widget.graph_widget.set_pixmap()
        return super().resizeEvent(event)

    @QtCore.Slot()
    def clear(self):
        self.composite_widget.graph_widget.figure.hide()
        self.composite_widget.table_widget.table.clear()
        self.arg_widget.get_symbol_input().clear()

class CorrelationWidget(QtWidgets.QWidget):
    def __init__(self, layer):
        super().__init__()
        self.setObjectName(layer)

    @staticmethod
    def _configure_control_skeleton():
        controls = generate_control_skeleton()
        controls['start_date'] = True
        controls['end_date'] = True
        return controls

    def _init_correlation_widgets(self):
        self.table_widget = TableWidget(widget_title="Correlation Matrix")
        self.arg_widget = ArgumentWidget(calculate_function=self.calculate,
                                            clear_function=self.clear,
                                            controls=CorrelationWidget._configure_control_skeleton(),
                                            layer=utilities.get_next_layer(self.objectName()))
        self.setLayout(QtWidgets.QVBoxLayout())

    def _arrange_correlation_widgets(self):
        self.layout().addWidget(self.table_widget)
        self.layout().addWidget(self.arg_widget)

    @QtCore.Slot()
    def calculate(self):
        if self.table_widget.table.isVisible():
            self.table_widget.table.clear()
            self.table_widget.table.hide()

        symbols = helper.split_and_strip(self.arg_widget.get_symbol_input().text())

        if len(symbols) > 1:
            self.table_widget.table.setRowCount(len(symbols))
            self.table_widget.table.setColumnCount(len(symbols))

            self.table_widget.table.setHorizontalHeaderLabels(symbols)
            self.table_widget.table.setVerticalHeaderLabels(symbols)

            matrix = statistics.correlation_matrix(tickers=symbols)
            for i in range(0, len(symbols)):
                for j in range(i, len(symbols)):
                    item_1 = QtWidgets.QTableWidgetItem(helper.format_float_percent(matrix[i][j]))
                    item_1.setTextAlignment(QtCore.Qt.AlignHCenter)

                    item_2 = QtWidgets.QTableWidgetItem(helper.format_float_percent(matrix[i][j]))
                    item_2.setTextAlignment(QtCore.Qt.AlignHCenter)

                    self.table_widget.table.setItem(j, i, item_1)
                    self.table_widget.table.setItem(i, j, item_2)
        else:
            print('error handling goes here')

        self.table_widget.show_table()

    @QtCore.Slot()
    def clear(self):
        self.arg_widget.get_symbol_input().clear()
        self.table_widget.table.clear()
        self.table_widget.table.hide()

class OptimizerWidget(QtWidgets.QWidget):
    def __init__(self, layer):
        super().__init__()
        self.setObjectName(layer)
        self._init_optimizer_widgets()
        self._arrange_optimizer_widgets()

    @staticmethod
    def _configure_control_skeleton():
        controls = generate_control_skeleton()
        controls['start_date'] = True
        controls['end_date'] = True
        return controls

    def _init_optimizer_widgets(self):
        self.table_widget = TableWidget(widget_title="Optimization Results")
        self.arg_widget = ArgumentWidget(calculate_function=self.optimize,
                                            clear_function=self.clear,
                                            controls=OptimizerWidget._configure_control_skeleton(),
                                            layer=utilities.get_next_layer(self.objectName()))
        self.setLayout(QtWidgets.QVBoxLayout())

    def _arrange_optimizer_widgets(self):
        self.layout().addWidget(self.table_widget)
        self.layout().addWidget(self.arg_widget)

    @QtCore.Slot()
    def optimize(self):
        if self.table_widget.table.isVisible():
            self.table_widget.table.clear()
            self.table_widget.table.hide()

        symbols = helper.split_and_strip(self.arg_widget.get_symbol_input().text())

        # TODO: better error checking
        if len(symbols) > 1:
            self.table_widget.table.setRowCount(len(symbols))
            self.table_widget.table.setVerticalHeaderLabels(symbols)

            # investment = self.portfolio_value.text()

            this_portfolio = Portfolio(tickers=symbols)
            allocation = optimizer.optimize_portfolio_variance(portfolio=this_portfolio)
            
            # self.result.setText(formats.format_allocation_profile_title(allocation, this_portfolio))
            
           # if not investment:
                # self.table_widget.table.setColumnCount(1)
                # labels = ['Allocation']
            # else:
                # self.table_widget.table.setColumnCount(2)
                # labels = ['Allocation', 'Shares']
                # shares = this_portfolio.calculate_approximate_shares(allocation, float(investment))

            # self.table_widget.table.setHorizontalHeaderLabels(labels)


            # for i in range(len(symbols)):
              #   item = QtWidgets.QTableWidgetItem(formats.format_allocation(allocation[i]))
                # item.setTextAlignment(QtCore.Qt.AlignHCenter)
                # self.table_widget.table.setItem(i, 0, item)
                # if investment:
                  #  share_item = QtWidgets.QTableWidgetItem(str(shares[i]))
                  #  share_item.setTextAlignment(QtCore.Qt.AlignHCenter)
                  #  self.table_widget.table.setItem(i, 1, share_item)
            
            # self.table_widget.table.resizeColumnsToContents()
            # self.table_widget.table.show()
            # self.result.show()
        
        # else:
          #   self.result.setText("Error Occurred. Check Input and Try Again.")
            # self.result.show()

    @QtCore.Slot()
    def clear(self):
        self.table_widget.table.clear()
        self.table_widget.table.hide()

class EfficientFrontierWidget(QtWidgets.QWidget):
    def __init__(self, layer):
        super().__init__()
        self.setObjectName(layer)
        self._init_frontier_widgets()
        self._arrange_frontier_widgets()
    @staticmethod
    def _configure_control_skeleton():
        controls = generate_control_skeleton()
        controls['start_date'] = True
        controls['end_date'] = True
        return controls

    def _init_frontier_widgets(self):
        self.graph_widget = GraphWidget(tmp_graph_key=keys.keys['GUI']['TEMP']['FRONTIER'])
        self.arg_widget = ArgumentWidget(calculate_function=self.calculate, 
                                            clear_function=self.clear,
                                            controls=EfficientFrontierWidget._configure_control_skeleton(),
                                            layer=utilities.get_next_layer(self.objectName()))
        self.setLayout(QtWidgets.QVBoxLayout())
        # TODO: portfolio tabs

    def _arrange_frontier_widgets(self):
        self.layout().addWidget(self.graph_widget)
        self.layout().addWidget(self.arg_widget)


    def resizeEvent(self, event: QtGui.QResizeEvent) -> None:
        if self.graph_widget.figure.isVisible():
            self.graph_widget.set_pixmap()
        return super().resizeEvent(event)

    @QtCore.Slot()
    def calculate(self):
        if self.graph_widget.figure.isVisible():
            self.graph_widget.figure.hide()

        symbols = helper.split_and_strip(self.arg_widget.get_symbol_input().text())

        this_portfolio = Portfolio(tickers=symbols)
        frontier = optimizer.calculate_efficient_frontier(portfolio=this_portfolio)
        plotter.plot_frontier(portfolio=this_portfolio, 
                                frontier=frontier, 
                                show=False, 
                                savefile=f'{settings.TEMP_DIR}/{keys.keys["GUI"]["TEMP"]["FRONTIER"]}')

        self.graph_widget.set_pixmap()
    
    @QtCore.Slot()
    def clear(self):
        self.graph_widget.figure.hide()
        self.arg_widget.get_symbol_input().clear()

class MovingAverageWidget(QtWidgets.QWidget):
    def __init__(self, layer):
        super().__init__()
        self.setObjectName(layer)
        self._init_average_widgets()
        self._arrange_average_widgets()

    def _init_average_widgets(self):
        self.graph_widget = GraphWidget(keys.keys['GUI']['TEMP']['AVERAGES'])
        self.arg_widget = ArgumentWidget(calculate_function=self.calculate,
                                            clear_function=self.clear,
                                            controls = generate_control_skeleton(),
                                            layer=utilities.get_next_layer(self.objectName()))
        self.setLayout(QtWidgets.QVBoxLayout())

    def _arrange_average_widgets(self):
        self.layout().addWidget(self.graph_widget)
        self.layout().addWidget(self.arg_widget)
        
    def resizeEvent(self, event: QtGui.QResizeEvent) -> None:
        if self.graph_widget.figure.isVisible():
            self.graph_widget.set_pixmap()
        return super().resizeEvent(event)

    @QtCore.Slot()
    def calculate(self):
        if self.graph_widget.figure.isVisible():
            self.graph_widget.figure.hide()

        symbols = helper.split_and_strip(self.arg_widget.get_symbol_input().text())

        moving_averages = statistics.calculate_moving_averages(symbols)
        periods = [settings.MA_1_PERIOD, settings.MA_2_PERIOD, settings.MA_3_PERIOD]
        plotter.plot_moving_averages(symbols=symbols, 
                                        averages_output=moving_averages, 
                                        periods=periods, 
                                        show=False, 
                                        savefile=f'{settings.TEMP_DIR}/averages')
        
        self.graph_widget.set_pixmap()
    
    @QtCore.Slot()
    def clear(self):
        self.arg_widget.get_symbol_input().clear()
        self.graph_widget.figure.hide()


import datetime
from typing import List

from scrilla.static import constants, formats, definitions


### GENERAL OUTPUT FUNCTIONS
def space(n: int):
    print('\n'*n)

def title_line(title: str, line_length: int=formats.formats['LINE_LENGTH'], separator: str=formats.formats['separator'],
display: bool=True):
    buff = int((line_length - len(title))/2)
    result = separator*buff + title + separator*buff
    if display:
        print(result)
    return result 
    
def separator_line(line_length=formats.formats['LINE_LENGTH'], separator=formats.formats['separator'], display=True) -> str:
    if display:
        print(separator*line_length)
    return separator*line_length

def break_lines(msg: str, line_length: int=formats.formats['LINE_LENGTH']) -> List[str]:
    """
    Generates a list of strings where each string is less than or equal to a specified line length. All elements in the array will be equal length except the last element; in other words, the inputted string will be read into an array, line by line, until there is nothing left to be read.

    Parameters
    ----------
    1. ``msg``: ``str``
        String to be broken into an array with each string no longer than `line_length`.
    2. **line_length**: ``int``
        *Optional*. Defaults to the value set by `static.formats.formats['LINE_LENGTH]`. The maximum length of a line. 
    
    Returns
    -------
    `list[str]`
        A list of `line`'s with `len(line)<line_length or len(line)==line_length`.
    """
    
    if len(msg)>line_length:
        return [msg[i:i+line_length] for i in range(0,len(msg), line_length)]
    return [msg]

def center(this_line: str, line_length: int=formats.formats['LINE_LENGTH'], display: bool=True) -> str:
    """
    Centers an inputted line relative to the line length. 

    Parameters
    ----------
    1. **this_line**: ``str``
        the string to centered.
    2. **line_length**: ``int``
        *Optional*. Defaults to the value set by `static.formats.formats['LINE_LENGTH]`. The length of the line that is being centered over. 
    3. **display**: ``bool``
        *Optional*. Defaults to `True`. If set to `True`, function will print result to screen. If set to `False`, the function will return the value as a string.
    
    Returns
    -------
    `str`
        The line with the difference between `line_length` and `this_line` distributed evenly on either side of the content as white space. 

    Raises
    ------
    1. **ValueError**
        If the inputted line is larged than the line length, this error will be thrown.
    """
    
    if len(this_line)>line_length:
        raise ValueError('Line to be centered is larger than length being centered over.')

    buff = int((line_length - len(this_line))/2)
    output = ' '*buff +  this_line + ' '*buff
    if display:
        print(output)
        return
    return output

def print_list(list_to_print, indent=formats.formats['INDENT']):
    for i, item in enumerate(list_to_print):
        print(' '*indent, f'{i}. {item}')

def string_result(operation, result, indent=formats.formats['INDENT'], display=True):
    output = ' '*indent, operation, ' = ', result
    if display:
        print(output)
        return
    return output

def scalar_result(calculation, result, currency=True, indent=formats.formats['INDENT']):
    if currency:
        print(' '*indent, calculation, ' = $', round(float(result), 2))
    else:
        print(' '*indent, calculation, ' = ', round(float(result), 4))

def percent_result(calculation, result, indent=formats.formats['INDENT']):
    print(' '*indent, calculation, ' = ', round(float(result), 4), '%')

def equivalent_result(right_hand, left_hand, value, indent=formats.formats['INDENT']):
    print(' '*indent, f'{right_hand} = {left_hand} = {value}')

def help_msg(indent=formats.formats['INDENT']):
    func_dict, arg_dict = definitions.FUNC_DICT, definitions.ARG_DICT

    title_line('scrilla')
    space(1)

    for paragraph in definitions.HELP_MSG:
        for line in break_lines(paragraph):
            center(line)
        space(1)

    title_line('SYNTAX')
    center(definitions.SYNTAX)
    space(1)

    for func_name in func_dict:
        title_line(func_dict[func_name]['name'])
        for line in break_lines(func_dict[func_name]['description']):
            center(line)
        separator_line()

        commands = func_dict[func_name]['values']
        print(' ', f'COMMAND: {commands[0]}, {commands[1]}')


        if func_dict[func_name]['args'] is not None:
            for arg_name in func_dict[func_name]['args']:
                aliases = arg_dict[arg_name]['values']

                print(' '*indent, f'OPTION: {aliases[0]}, {aliases[1]}, {aliases[2]}, {aliases[3]}')

                if arg_dict[arg_name]['required']:
                    print(' '*2*indent, 'REQUIRED')

                print(' '*2*indent, f'NAME: {arg_dict[arg_name]["name"]}')

                if arg_dict[arg_name]['default'] is not None:
                    print(' '*2*indent, f'DEFAULT: {arg_dict[arg_name]["default"]}')
                    
                if arg_dict[arg_name]['syntax'] is not None:
                    print(' '*2*indent, f'FORMAT: {arg_dict[arg_name]["syntax"]}')    
        separator_line()    

### ANALYSIS SPECIFIC OUTPUT FUNCTIONS
def portfolio_percent_result(result, tickers,indent=formats.formats['INDENT']):
    for i, item in enumerate(tickers):
        print(' '*indent, f'{item} =', round(100*result[i], 2), '%')

def portfolio_shares_result(result, tickers, indent=formats.formats['INDENT']):
    for i, item in enumerate(tickers):
        print(' '*indent, f'{item} =', result[i])

def spot_price(ticker, this_spot_price):
    formatted_price = round(float(this_spot_price), 2)
    scalar_result(f'{ticker} spot price', formatted_price)
    
def model_price(ticker, this_model_price, model):
    formatted_price = round(float(this_model_price),2)
    scalar_result(f'{ticker} {str(model).upper()} price', formatted_price)

def risk_profile(profiles):
    for key, value in profiles.items():
        title_line(f'{key} Risk Profile')
        for subkey, subvalue in value.items():
            scalar_result(f'{subkey}', f'{subvalue}', currency=False)

def moving_average_result(tickers, averages_output, periods, start_date = None, end_date = None):
    averages, dates = averages_output
    MA1_prefix, MA2_prefix, MA3_prefix = f'MA({periods[0]})', f'MA({periods[1]})', f'MA({periods[2]})'
    if start_date is None and end_date is None:
        for i, item in enumerate(tickers):
            title = f'{item} Moving Average of Daily Return for {periods[0]}, {periods[1]} & {periods[0]} Days'
            title_line(title)

            MA1_title, MA2_title, MA3_title = f'{MA1_prefix}_{item}', f'{MA2_prefix}_{item}', f'{MA3_prefix}_{item}'
            scalar_result(MA1_title, round(averages[i][0], 2))
            scalar_result(MA2_title, round(averages[i][1], 2))
            scalar_result(MA3_title, round(averages[i][2], 2))
    else:
        for i, item in enumerate(tickers):

            title = f'{item} Moving Average of Daily Return for {periods[0]}, {periods[1]} & {periods[0]} Days'
            title_line(title)

            MA1_title, MA2_title, MA3_title = f'{MA1_prefix}_{item}', f'{MA2_prefix}_{item}', f'{MA3_prefix}_{item}'
            for j, item in enumerate(dates):
                msg_1 = f'{item} : {MA1_title}'
                scalar_result(msg_1, round(averages[i][0][j], 2))
            for j, item in enumerate(dates):
                msg_2 = f'{item} : {MA2_title}'
                scalar_result(msg_2, round(averages[i][1][j], 2))
            for j, item in enumerate(dates):  
                msg_3 = f'{item} : {MA3_title}'
                scalar_result(msg_3, round(averages[i][2][j], 2))      

def screen_results(info, model):
    for ticker in info:
        title_line(f'{ticker} {str(model).upper()} Model vs. Spot Price ')
        spot_price(ticker=ticker, this_spot_price=info[ticker]['spot_price'])
        model_price(ticker=ticker, this_model_price=info[ticker]['model_price'], model=model)
        scalar_result(f'{ticker} discount', info[ticker]['discount'])
        separator_line()

# TODO: can probably combine optimal_result and efficient_frontier into a single function
#         by wrapping the optimal_results in an array so when it iterates through frontier
#         in efficient_frontier, it will only pick up the single allocation array for the
#         optimal result.
def optimal_result(portfolio, allocation, investment=None):
    title_line('Optimal Percentage Allocation')
    portfolio_percent_result(allocation, portfolio.tickers)

    if investment is not None:
        shares = portfolio.calculate_approximate_shares(allocation, investment)
        total = portfolio.calculate_actual_total(allocation, investment)
        
        title_line('Optimal Share Allocation')
        portfolio_shares_result(shares, portfolio.tickers)
        title_line('Optimal Portfolio Value')
        scalar_result('Total', round(total,2))

    title_line('Risk-Return Profile')
    scalar_result(calculation='Return', result=portfolio.return_function(allocation), currency=False)
    scalar_result(calculation='Volatility', result=portfolio.volatility_function(allocation), currency=False)

def efficient_frontier(portfolio, frontier, investment=None):
    title_line('(Annual Return %, Annual Volatility %) Portfolio')

    # TODO: edit title to include dates

    for allocation in frontier:
        separator_line()
        return_string=str(round(round(portfolio.return_function(allocation),4)*100,2))
        vol_string=str(round(round(portfolio.volatility_function(allocation),4)*100,2))
        title_line(f'({return_string} %, {vol_string}%) Portfolio')
        separator_line()

        title_line('Optimal Percentage Allocation')
        portfolio_percent_result(allocation, portfolio.tickers)
        
        if investment is not None:
            shares = portfolio.calculate_approximate_shares(allocation, investment)
            total = portfolio.calculate_actual_total(allocation, investment)
        
            title_line('Optimal Share Allocation')
            portfolio_shares_result(shares, portfolio.tickers)
            title_line('Optimal Portfolio Value')
            scalar_result('Total', round(total,2))
        
        title_line('Risk-Return Profile')
        scalar_result('Return', portfolio.return_function(allocation), currency=False)
        scalar_result('Volatility', portfolio.volatility_function(allocation), currency=False)
        print('\n')

def correlation_matrix(tickers: List[str], correlation_matrix: List[List[float]], display: bool = True):
    """
    Parameters
    ----------
    1. **tickers**: ``list``
        Array of tickers for which the correlation matrix was calculated and formatted.
    2. **indent**: ``int``
        Amount of indent on each new line of the correlation matrix.
    3. **start_date**: ``datetime.date`` 
        Start date of the time period over which correlation was calculated. 
    4. **end_date**: ``datetime.date`` 
        End date of the time period over which correlation was calculated. 
    
    Returns
    ------
    A correlation matrix string formatted with new lines and spaces.
    """
    entire_formatted_result, formatted_title = "", ""

    line_length, first_symbol_length = 0, 0
    new_line=""
    no_symbols = len(tickers)

    for i in range(no_symbols):
        this_symbol = tickers[i]
        symbol_string = ' '*formats.formats['INDENT'] + f'{this_symbol} '

        if i != 0:
            this_line = symbol_string + ' '*(line_length - len(symbol_string) - 7*(no_symbols - i))
            # NOTE: seven is number of chars in ' 100.0%'
        else: 
            this_line = symbol_string
            first_symbol_length = len(this_symbol)

        new_line = this_line
        
        for j in range(i, no_symbols):
            if j == i:
                new_line += " 100.0%"
            
            else:
                result = correlation_matrix[i][j]
                formatted_result = str(100*result)[:constants.constants['SIG_FIGS']]
                new_line += f' {formatted_result}%'

        entire_formatted_result += new_line + '\n'
        
        if i == 0:
            line_length = len(new_line)

    formatted_title += ' '*(formats.formats['INDENT'] + first_symbol_length+1)
    for symbol in tickers:
        sym_len = len(symbol)
        formatted_title += f' {symbol}'+ ' '*(7-sym_len)
        # NOTE: seven is number of chars in ' 100.0%'
    formatted_title += '\n'

    whole_thing = formatted_title + entire_formatted_result

    if display:
        print(f'\n{whole_thing}')
        return

    return whole_thing

class Logger():

    def __init__(self, location, log_level="info"):
        self.location = location
        self.log_level = log_level
    
    # LOGGING FUNCTIONS
    def comment(self, msg):
        now = datetime.datetime.now()
        dt_string = now.strftime("%d/%m/%Y %H:%M:%S")
        print(dt_string, ' :' , self.location, ' : ',msg)

    def info(self, msg):
        if self.log_level in [constants.constants['LOG_LEVEL']['INFO'], 
                                constants.constants['LOG_LEVEL']['DEBUG'], 
                                constants.constants['LOG_LEVEL']['VERBOSE']]:
            self.comment(msg)

    def debug(self, msg):
        if self.log_level in [constants.constants['LOG_LEVEL']['DEBUG'], 
                                constants.constants['LOG_LEVEL']['VERBOSE']]:
            self.comment(msg)

    def verbose(self, msg):
        if self.log_level == constants.constants['LOG_LEVEL']['VERBOSE']:
            self.comment(msg)
